"""Built in actions for Jaseci"""
from operator import itemgetter
from jaseci.utils.utils import app_logger, json_out
from datetime import datetime
from jaseci.jac.machine.jac_value import jac_wrap_value as jwv
from jaseci.actions.live_actions import jaseci_action
from jaseci.element.element import element
import sys
import uuid
import json


@jaseci_action()
def log(*args):
    """Standard built in for printing output to log"""
    result = ''
    for i in args:
        result += json_out(jwv(i))
    app_logger.info(result)
    return result


@jaseci_action()
def out(*args):
    """Standard built in for printing output"""
    args = [json_out(jwv(x)) for x in args]
    print(*args)


@jaseci_action()
def err(*args):
    """Standard built in for printing to stderr"""
    args = [json_out(jwv(x)) for x in args]
    print(*args, file=sys.stderr)


@jaseci_action()
def sort_by_col(lst: list, col_num: int, reverse: bool = False):
    """
    Sorts in place list of lists by column
    Param 1 - list
    Param 2 - col number
    Param 3 - boolean as to whether things should be reversed

    Return - Sorted list
    """
    return sorted(lst, key=itemgetter(col_num), reverse=reverse)


@jaseci_action()
def time_now(meta):
    """Get utc date time for now in iso format"""
    return datetime.utcnow().isoformat()


@jaseci_action()
def set_global(name: str, value, meta):
    """
    Set global variable visible to all walkers/users
    Param 1 - name
    Param 2 - value (must be json serializable)
    """
    mast = meta['h'].get_obj(meta['m_id'], uuid.UUID(meta['m_id']))
    if(not mast.is_master(super_check=True, silent=False)):
        return False
    mast.global_set(name, json.dumps(value))
    return json.loads(mast.global_get(name)['value'])


@jaseci_action()
def get_global(name: str, meta):
    """
    Set global variable visible to all walkers/users
    Param 1 - name
    """
    mast = meta['h'].get_obj(meta['m_id'], uuid.UUID(meta['m_id']))
    val = mast.global_get(name)['value']
    if(val):
        return json.loads(val)
    else:
        return None


@jaseci_action()
def destroy_global(name: str, meta):
    """Get utc date time for now in iso format"""
    mast = meta['h'].get_obj(meta['m_id'], uuid.UUID(meta['m_id']))
    if(not mast.is_master(super_check=True, silent=False)):
        return False
    return mast.global_delete(name)


@jaseci_action()
def set_perms(obj: element, mode: str, meta):
    """
    Sets object access mode for any Jaseci object
    Param 1 - target element
    Param 2 - valid permission (public, private, read_only)

    Return - true/false whether successful
    """
    mast = meta['h'].get_obj(meta['m_id'], uuid.UUID(meta['m_id']))
    return mast.object_perms_set(obj=obj,
                                 mode=mode)['success']


@jaseci_action()
def get_perms(obj: element):
    """
    Returns object access mode for any Jaseci object
    Param 1 - target element

    Return - Sorted list
    """
    return obj.j_access


@jaseci_action()
def grant_perms(obj: element, mast: element, read_only: bool, meta):
    """
    Grants another user permissions to access a Jaseci object
    Param 1 - target element
    Param 2 - master to be granted permission
    Param 3 - Boolean read_only flag

    Return - Sorted list
    """
    mast = meta['h'].get_obj(meta['m_id'], uuid.UUID(meta['m_id']))
    return mast.object_perms_grant(obj=obj,
                                   mast=mast,
                                   read_only=read_only)['success']


@jaseci_action()
def revoke_perms(obj: element, mast: element, meta):
    """
    Remove permissions for user to access a Jaseci object
    Param 1 - target element
    Param 2 - master to be revoked permission

    Return - Sorted list
    """
    mast = meta['h'].get_obj(meta['m_id'], uuid.UUID(meta['m_id']))
    return mast.object_perms_revoke(obj=obj,
                                    mast=mast)['success']


@jaseci_action()
def get_report(meta):
    """
    Get current report so far for walker run
    """
    return meta['interp'].report




from jaseci.utils.mem_hook import mem_hook
from jaseci.actor.sentinel import sentinel
from jaseci.graph.graph import graph
from jaseci.jac.machine.jac_value import jac_elem_unwrap as jeu

from jaseci.utils.utils import TestCaseHelper
from unittest import TestCase
import jaseci.tests.jac_test_code as jtc


class jac_tests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()
        self.gph = graph(m_id='anon', h=mem_hook())
        self.sent = sentinel(m_id='anon', h=self.gph._h)

    def tearDown(self):
        super().tearDown()

    def test_ll_proto_load(self):
        """Test loading/parsing ll prototype"""
        self.sent.register_code(jtc.ll_proto)
        self.assertTrue(self.sent.is_active)

    def test_rand_generation(self):
        """Test loading/parsing ll prototype"""
        self.sent.register_code(jtc.ll_proto)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        gen_walker = self.sent.walker_ids.get_obj_by_name(
            'gen_rand_life', kind='walker')
        gen_walker.prime(self.gph.outbound_nodes()[0])
        gen_walker.run()
        self.assertGreater(len(self.gph._h.mem.keys()), 70)

    def test_objects_created(self):
        """Test loading/parsing ll prototype"""
        self.sent.register_code(jtc.ll_proto)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        gen_walker = self.sent.walker_ids.get_obj_by_name(
            'gen_rand_life', kind='walker')
        gen_walker.prime(self.gph.outbound_nodes()[0])
        gen_walker.run()
        for i in self.gph._h.mem.keys():
            if(i == 'global'):
                continue
            self.gph._h.mem[i].json()

    def test_get_latest_day(self):
        """Test loading/parsing ll prototype"""
        self.sent.register_code(jtc.ll_proto)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        gen_walker = self.sent.walker_ids.get_obj_by_name(
            'gen_rand_life', kind='walker')
        gen_walker.prime(self.gph.outbound_nodes()[0])
        gen_walker.run()
        lday_walk = self.sent.walker_ids.get_obj_by_name(
            'get_latest_day')
        lday_walk.prime(self.gph.outbound_nodes()[0])
        lday_walk.run()
        ret = lday_walk.context['latest_day']
        self.assertEqual(jeu(ret, self.gph).name, 'day')

    def test_carry_forward(self):
        """Test loading/parsing ll prototype"""
        self.sent.register_code(jtc.ll_proto)
        gen_walker = self.sent.walker_ids.get_obj_by_name('init')
        gen_walker.prime(self.gph)
        gen_walker.run()
        gen_walker = self.sent.walker_ids.get_obj_by_name(
            'gen_rand_life', kind='walker')
        gen_walker.prime(self.gph)
        gen_walker.run()
        lday_walk = self.sent.walker_ids.get_obj_by_name(
            'get_gen_day')
        lday_walk.prime(self.gph)
        lday_walk.run()
        day = jeu(lday_walk.context['day_node'], self.gph)
        self.assertGreater(len(day.edge_ids), 3)



"""
This module includes code related to hooking Jaseci's Redis to the
core engine.
"""
from jaseci.utils import utils
from jaseci.utils.utils import logger
import jaseci as core_mod
from jaseci.utils.mem_hook import mem_hook
from redis import Redis
import json


class redis_hook(mem_hook):
    """
    Hooks Django ORM database for Jaseci objects to Jaseci's core engine.

    Sets user on initialization to route calls to correct user. This hook
    lives in :class:`User` class as per :field:`User.orm_hook`.
    """

    def __init__(self, red=Redis()):
        self.red = red
        self.save_obj_list = set()
        self.save_glob_dict = {}
        super().__init__()

    def get_obj_from_store(self, item_id):
        loaded_obj = self.red.get(item_id.urn)
        if(not loaded_obj):
            logger.error(
                str(f"Object {item_id} does not exist in Redis!")
            )
            return None

        jdict = json.loads(loaded_obj)
        j_type = jdict['j_type']
        j_master = jdict['j_master']
        class_for_type = \
            utils.find_class_and_import(j_type, core_mod)
        ret_obj = class_for_type(h=self, m_id=j_master, auto_save=False)
        ret_obj.json_load(loaded_obj)

        return ret_obj

    def has_obj_in_store(self, item_id):
        """
        Checks for object existance in store
        """
        if (self.red.exists(item_id.urn)):
            return True
        return False

    def save_obj_to_store(self, item):
        self.save_obj_list.add(item)

    def commit_obj(self, item):
        try:
            self.red.set(item.id.urn, item.json(detailed=True))
        except TypeError:
            logger.error(
                str(f"Item {item} is not JSON serializable for redis store!"),
                exc_info=True
            )
        except Exception as e:
            logger.error(
                str(f"Couldn't save {item} to redis! {e}"),
                exc_info=True
            )

    def destroy_obj_from_store(self, item):
        self.red.delete(item.id.urn)
        if(item in self.save_obj_list):
            self.save_obj_list.remove(item)

    def get_glob_from_store(self, name):
        """
        Get global config from externally hooked general store by name
        """
        loaded_val = self.red.get(name)
        if (loaded_val):
            return loaded_val
        else:
            logger.error(
                str(f"Global {name} does not exist in Django ORM!"),
                exc_info=True
            )
            return None

    def has_glob_in_store(self, name):
        """
        Checks for global config existance in store
        """
        if (self.red.exists(name)):
            return True
        return False

    def save_glob_to_store(self, name, value):
        """Save global config to externally hooked general store"""
        self.save_glob_dict[name] = value

    def list_glob_from_store(self):
        """Get list of globals to externally hooked general store"""
        logger.warning(str(f"Globals can not (yet) be listed from Redis!"))
        return []

    def destroy_glob_from_store(self, name):
        """Destroy global config to externally hooked general store"""
        self.red.delete(name)
        if(name in self.save_glob_dict.keys()):
            del self.save_glob_dict[name]

    def commit_glob(self, name, value):
        try:
            self.red.set(name, value)
        except Exception as e:
            logger.error(
                str(f"Couldn't save {name} to redis! {e}"),
                exc_info=True
            )

    def commit(self):
        """Write through all saves to store"""
        for i in self.save_obj_list:
            self.commit_obj(i)
        self.save_obj_list = set()
        for i in self.save_glob_dict.keys():
            self.commit_glob(name=i, value=self.save_glob_dict[i])
        self.save_glob_dict = {}



"""
Sentinel interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.actor.architype import architype
from jaseci.actor.walker import walker
from jaseci.jac.interpreter.interp import interp
from jaseci.utils.utils import parse_str_token
from jaseci.jac.ir.jac_code import jac_ast_to_ir


class sentinel_interp(interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    def run_start(self, jac_ast):
        """
        start: ver_label? element+ EOF;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'ver_label'):
            self.run_ver_label(kid[0])
        for i in kid[:-1]:
            self.run_element(i)

    def run_ver_label(self, jac_ast):
        """
        ver_label: 'version' COLON STRING;
        """
        kid = self.set_cur_ast(jac_ast)
        self.version = parse_str_token(kid[2].token_text())

    def run_element(self, jac_ast):
        """
        element: architype | walker | test;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'architype'):
            self.load_architype(kid[0])
        elif(kid[0].name == 'walker'):
            self.load_walker(kid[0])
        elif(kid[0].name == 'test'):
            self.load_test(kid[0])

    def load_architype(self, jac_ast):
        """
        architype:
            KW_NODE NAME (COLON INT)? attr_block
            | KW_EDGE NAME attr_block
            | KW_GRAPH NAME graph_block;
        """
        arch = architype(m_id=self._m_id, h=self._h, code_ir=jac_ast)
        if(self.arch_ids.has_obj_by_name(arch.name, kind=arch.kind)):
            self.arch_ids.destroy_obj_by_name(arch.name, kind=arch.kind)
        self.arch_ids.add_obj(arch)
        return arch

    # Note: Sentinels only registers the attr_stmts
    def load_walker(self, jac_ast):
        """
        walker: KW_WALKER NAME namespaces? walker_block;
        """
        walk = walker(m_id=self._m_id, h=self._h, code_ir=jac_ast)
        if(jac_ast.kid[2].name == 'namespaces'):
            walk.namespaces = self.run_namespaces(jac_ast.kid[2])
        if(self.walker_ids.has_obj_by_name(walk.name)):
            self.walker_ids.destroy_obj_by_name(walk.name)
        self.walker_ids.add_obj(walk)
        return walk

    def load_test(self, jac_ast):
        """
        test:
            KW_TEST STRING KW_WITH (graph_ref | KW_GRAPH graph_block) KW_BY (
                (walker_ref spawn_ctx? (code_block | SEMI))
                | KW_WALKER walker_block
            );
        """
        kid = self.set_cur_ast(jac_ast)
        testcase = {'title': kid[1].token_text(),
                    'graph_ref': None, 'graph_block': None,
                    'walker_ref': None, 'spawn_ctx': None,
                    'assert_block': None, 'walker_block': None,
                    'outcome': None}
        kid = kid[3:]
        if(kid[0].name == "graph_ref"):
            graph_name = kid[0].kid[2].token_text()
            if(not self.arch_ids.has_obj_by_name(graph_name,
                                                 kind='graph')):
                self.rt_error(f"Graph {graph_name} not found!", kid[0])
                return
            testcase['graph_ref'] = graph_name
        else:
            kid = kid[1:]
            testcase['graph_block'] = jac_ast_to_ir(kid[0])
        kid = kid[2:]
        if(kid[0].name == 'walker_ref'):
            walker_name = kid[0].kid[2].token_text()
            if(not self.walker_ids.has_obj_by_name(walker_name)):
                self.rt_error(f"Walker {walker_name} not found!", kid[0])
                return
            testcase['walker_ref'] = walker_name
            kid = kid[1:]
            if(kid[0].name == 'spawn_ctx'):
                testcase['spawn_ctx'] = jac_ast_to_ir(kid[0])
                kid = kid[1:]
            if(kid[0].name == "code_block"):
                testcase['assert_block'] = jac_ast_to_ir(kid[0])
        else:
            kid = kid[1:]
            testcase['walker_block'] = jac_ast_to_ir(kid[0])

        self.testcases.append(testcase)

    def run_namespaces(self, jac_ast):
        """
        namespaces: COLON name_list;
        """
        return self.run_name_list(jac_ast.kid[1])



"""Built in actions for Jaseci"""
import random
# import faker
from datetime import datetime
from datetime import timedelta
from jaseci.actions.live_actions import jaseci_action


@jaseci_action()
def seed(val: int):
    """Seed random num generator"""
    random.seed(val, version=2)
    # faker.Faker.seed(param_list[0])


@jaseci_action()
def integer(start: int, end: int):
    """Random integeter between range"""
    return random.randint(start, end)


@jaseci_action()
def sentence(meta):
    """Get a random sentence"""
    fstr = ''
    for i in range(random.randint(0, 10)):
        fstr += "test "
    return fstr
    # return faker.Faker().sentence()


@jaseci_action()
def time(start_date: str, end_date: str):
    """Provide a random datetime between range"""
    start = datetime.fromisoformat(start_date)
    end = datetime.fromisoformat(end_date)
    return (start + timedelta(
        seconds=random.randint(0, int((end - start).total_seconds())),
    )).isoformat()



"""
Architype class for Jaseci

Each architype is a registered templatized version of instances of any Jaseci
abstractions or collections of instances (e.g., subgraphs, etc)
"""
from jaseci.element.element import element
from jaseci.jac.interpreter.architype_interp import architype_interp
from jaseci.jac.ir.jac_code import jac_code


class architype(element, jac_code, architype_interp):
    """Architype class for Jaseci"""

    def __init__(self, code_ir=None, *args, **kwargs):
        element.__init__(self, *args, **kwargs)
        jac_code.__init__(self, code_ir)
        architype_interp.__init__(self)

    def run(self):
        """
        Create set of new object instances from architype if needed
        """
        if(not self._jac_ast):
            self.refresh()
        return self.run_architype(jac_ast=self._jac_ast)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        super().destroy()



"""
Action class for Jaseci

Each action has an id, name, timestamp and it's set of edges.
"""
from .item import item
from jaseci.actions.live_actions import live_actions
import inspect
# ACTION_PACKAGE = 'jaseci.actions.'


class action(item):
    """
    Action class for Jaseci

    preset_in_out holds a set of parameters in the form of lists of context
    objects that are to be used from whereever those contexts are attached
    e.g., (nodes, edges, walkers, etc). This is used by Jac's runtime
    engine to support preset actions in nodes
    access_list is used by walker to decide what to trigger
    """

    def __init__(self, preset_in_out=None, access_list=None,
                 *args, **kwargs):
        self.preset_in_out = preset_in_out  # Not using _ids convention
        self.access_list = access_list
        super().__init__(*args, **kwargs)

    def trigger(self, param_list, scope, interp):
        """
        param_list should be passed as list of values to lib functions
        Also note that Jac stores preset_in_out as input/output list of hex
        ids since preset_in_out doesn't use _ids convention
        """
        func = live_actions[self.value]
        args = inspect.getfullargspec(func)
        args = args[0]+args[4]
        if('meta' in args):
            result = func(*param_list,
                          meta={'m_id': scope.parent._m_id,
                                'h': scope.parent._h, 'scope': scope,
                                'interp':  interp})
        else:
            result = func(*param_list)
        return result



# Generated from jac.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .jacParser import jacParser
else:
    from jacParser import jacParser

# This class defines a complete listener for a parse tree produced by jacParser.
class jacListener(ParseTreeListener):

    # Enter a parse tree produced by jacParser#start.
    def enterStart(self, ctx:jacParser.StartContext):
        pass

    # Exit a parse tree produced by jacParser#start.
    def exitStart(self, ctx:jacParser.StartContext):
        pass


    # Enter a parse tree produced by jacParser#import_module.
    def enterImport_module(self, ctx:jacParser.Import_moduleContext):
        pass

    # Exit a parse tree produced by jacParser#import_module.
    def exitImport_module(self, ctx:jacParser.Import_moduleContext):
        pass


    # Enter a parse tree produced by jacParser#ver_label.
    def enterVer_label(self, ctx:jacParser.Ver_labelContext):
        pass

    # Exit a parse tree produced by jacParser#ver_label.
    def exitVer_label(self, ctx:jacParser.Ver_labelContext):
        pass


    # Enter a parse tree produced by jacParser#import_items.
    def enterImport_items(self, ctx:jacParser.Import_itemsContext):
        pass

    # Exit a parse tree produced by jacParser#import_items.
    def exitImport_items(self, ctx:jacParser.Import_itemsContext):
        pass


    # Enter a parse tree produced by jacParser#import_names.
    def enterImport_names(self, ctx:jacParser.Import_namesContext):
        pass

    # Exit a parse tree produced by jacParser#import_names.
    def exitImport_names(self, ctx:jacParser.Import_namesContext):
        pass


    # Enter a parse tree produced by jacParser#element.
    def enterElement(self, ctx:jacParser.ElementContext):
        pass

    # Exit a parse tree produced by jacParser#element.
    def exitElement(self, ctx:jacParser.ElementContext):
        pass


    # Enter a parse tree produced by jacParser#architype.
    def enterArchitype(self, ctx:jacParser.ArchitypeContext):
        pass

    # Exit a parse tree produced by jacParser#architype.
    def exitArchitype(self, ctx:jacParser.ArchitypeContext):
        pass


    # Enter a parse tree produced by jacParser#walker.
    def enterWalker(self, ctx:jacParser.WalkerContext):
        pass

    # Exit a parse tree produced by jacParser#walker.
    def exitWalker(self, ctx:jacParser.WalkerContext):
        pass


    # Enter a parse tree produced by jacParser#walker_block.
    def enterWalker_block(self, ctx:jacParser.Walker_blockContext):
        pass

    # Exit a parse tree produced by jacParser#walker_block.
    def exitWalker_block(self, ctx:jacParser.Walker_blockContext):
        pass


    # Enter a parse tree produced by jacParser#test.
    def enterTest(self, ctx:jacParser.TestContext):
        pass

    # Exit a parse tree produced by jacParser#test.
    def exitTest(self, ctx:jacParser.TestContext):
        pass


    # Enter a parse tree produced by jacParser#namespaces.
    def enterNamespaces(self, ctx:jacParser.NamespacesContext):
        pass

    # Exit a parse tree produced by jacParser#namespaces.
    def exitNamespaces(self, ctx:jacParser.NamespacesContext):
        pass


    # Enter a parse tree produced by jacParser#walk_entry_block.
    def enterWalk_entry_block(self, ctx:jacParser.Walk_entry_blockContext):
        pass

    # Exit a parse tree produced by jacParser#walk_entry_block.
    def exitWalk_entry_block(self, ctx:jacParser.Walk_entry_blockContext):
        pass


    # Enter a parse tree produced by jacParser#walk_exit_block.
    def enterWalk_exit_block(self, ctx:jacParser.Walk_exit_blockContext):
        pass

    # Exit a parse tree produced by jacParser#walk_exit_block.
    def exitWalk_exit_block(self, ctx:jacParser.Walk_exit_blockContext):
        pass


    # Enter a parse tree produced by jacParser#walk_activity_block.
    def enterWalk_activity_block(self, ctx:jacParser.Walk_activity_blockContext):
        pass

    # Exit a parse tree produced by jacParser#walk_activity_block.
    def exitWalk_activity_block(self, ctx:jacParser.Walk_activity_blockContext):
        pass


    # Enter a parse tree produced by jacParser#attr_block.
    def enterAttr_block(self, ctx:jacParser.Attr_blockContext):
        pass

    # Exit a parse tree produced by jacParser#attr_block.
    def exitAttr_block(self, ctx:jacParser.Attr_blockContext):
        pass


    # Enter a parse tree produced by jacParser#attr_stmt.
    def enterAttr_stmt(self, ctx:jacParser.Attr_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#attr_stmt.
    def exitAttr_stmt(self, ctx:jacParser.Attr_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#graph_block.
    def enterGraph_block(self, ctx:jacParser.Graph_blockContext):
        pass

    # Exit a parse tree produced by jacParser#graph_block.
    def exitGraph_block(self, ctx:jacParser.Graph_blockContext):
        pass


    # Enter a parse tree produced by jacParser#graph_block_spawn.
    def enterGraph_block_spawn(self, ctx:jacParser.Graph_block_spawnContext):
        pass

    # Exit a parse tree produced by jacParser#graph_block_spawn.
    def exitGraph_block_spawn(self, ctx:jacParser.Graph_block_spawnContext):
        pass


    # Enter a parse tree produced by jacParser#graph_block_dot.
    def enterGraph_block_dot(self, ctx:jacParser.Graph_block_dotContext):
        pass

    # Exit a parse tree produced by jacParser#graph_block_dot.
    def exitGraph_block_dot(self, ctx:jacParser.Graph_block_dotContext):
        pass


    # Enter a parse tree produced by jacParser#has_root.
    def enterHas_root(self, ctx:jacParser.Has_rootContext):
        pass

    # Exit a parse tree produced by jacParser#has_root.
    def exitHas_root(self, ctx:jacParser.Has_rootContext):
        pass


    # Enter a parse tree produced by jacParser#has_stmt.
    def enterHas_stmt(self, ctx:jacParser.Has_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#has_stmt.
    def exitHas_stmt(self, ctx:jacParser.Has_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#has_assign.
    def enterHas_assign(self, ctx:jacParser.Has_assignContext):
        pass

    # Exit a parse tree produced by jacParser#has_assign.
    def exitHas_assign(self, ctx:jacParser.Has_assignContext):
        pass


    # Enter a parse tree produced by jacParser#can_stmt.
    def enterCan_stmt(self, ctx:jacParser.Can_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#can_stmt.
    def exitCan_stmt(self, ctx:jacParser.Can_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#event_clause.
    def enterEvent_clause(self, ctx:jacParser.Event_clauseContext):
        pass

    # Exit a parse tree produced by jacParser#event_clause.
    def exitEvent_clause(self, ctx:jacParser.Event_clauseContext):
        pass


    # Enter a parse tree produced by jacParser#preset_in_out.
    def enterPreset_in_out(self, ctx:jacParser.Preset_in_outContext):
        pass

    # Exit a parse tree produced by jacParser#preset_in_out.
    def exitPreset_in_out(self, ctx:jacParser.Preset_in_outContext):
        pass


    # Enter a parse tree produced by jacParser#dotted_name.
    def enterDotted_name(self, ctx:jacParser.Dotted_nameContext):
        pass

    # Exit a parse tree produced by jacParser#dotted_name.
    def exitDotted_name(self, ctx:jacParser.Dotted_nameContext):
        pass


    # Enter a parse tree produced by jacParser#name_list.
    def enterName_list(self, ctx:jacParser.Name_listContext):
        pass

    # Exit a parse tree produced by jacParser#name_list.
    def exitName_list(self, ctx:jacParser.Name_listContext):
        pass


    # Enter a parse tree produced by jacParser#expr_list.
    def enterExpr_list(self, ctx:jacParser.Expr_listContext):
        pass

    # Exit a parse tree produced by jacParser#expr_list.
    def exitExpr_list(self, ctx:jacParser.Expr_listContext):
        pass


    # Enter a parse tree produced by jacParser#code_block.
    def enterCode_block(self, ctx:jacParser.Code_blockContext):
        pass

    # Exit a parse tree produced by jacParser#code_block.
    def exitCode_block(self, ctx:jacParser.Code_blockContext):
        pass


    # Enter a parse tree produced by jacParser#node_ctx_block.
    def enterNode_ctx_block(self, ctx:jacParser.Node_ctx_blockContext):
        pass

    # Exit a parse tree produced by jacParser#node_ctx_block.
    def exitNode_ctx_block(self, ctx:jacParser.Node_ctx_blockContext):
        pass


    # Enter a parse tree produced by jacParser#statement.
    def enterStatement(self, ctx:jacParser.StatementContext):
        pass

    # Exit a parse tree produced by jacParser#statement.
    def exitStatement(self, ctx:jacParser.StatementContext):
        pass


    # Enter a parse tree produced by jacParser#if_stmt.
    def enterIf_stmt(self, ctx:jacParser.If_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#if_stmt.
    def exitIf_stmt(self, ctx:jacParser.If_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#try_stmt.
    def enterTry_stmt(self, ctx:jacParser.Try_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#try_stmt.
    def exitTry_stmt(self, ctx:jacParser.Try_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#else_from_try.
    def enterElse_from_try(self, ctx:jacParser.Else_from_tryContext):
        pass

    # Exit a parse tree produced by jacParser#else_from_try.
    def exitElse_from_try(self, ctx:jacParser.Else_from_tryContext):
        pass


    # Enter a parse tree produced by jacParser#elif_stmt.
    def enterElif_stmt(self, ctx:jacParser.Elif_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#elif_stmt.
    def exitElif_stmt(self, ctx:jacParser.Elif_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#else_stmt.
    def enterElse_stmt(self, ctx:jacParser.Else_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#else_stmt.
    def exitElse_stmt(self, ctx:jacParser.Else_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#for_stmt.
    def enterFor_stmt(self, ctx:jacParser.For_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#for_stmt.
    def exitFor_stmt(self, ctx:jacParser.For_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#while_stmt.
    def enterWhile_stmt(self, ctx:jacParser.While_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#while_stmt.
    def exitWhile_stmt(self, ctx:jacParser.While_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#ctrl_stmt.
    def enterCtrl_stmt(self, ctx:jacParser.Ctrl_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#ctrl_stmt.
    def exitCtrl_stmt(self, ctx:jacParser.Ctrl_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#assert_stmt.
    def enterAssert_stmt(self, ctx:jacParser.Assert_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#assert_stmt.
    def exitAssert_stmt(self, ctx:jacParser.Assert_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#destroy_action.
    def enterDestroy_action(self, ctx:jacParser.Destroy_actionContext):
        pass

    # Exit a parse tree produced by jacParser#destroy_action.
    def exitDestroy_action(self, ctx:jacParser.Destroy_actionContext):
        pass


    # Enter a parse tree produced by jacParser#report_action.
    def enterReport_action(self, ctx:jacParser.Report_actionContext):
        pass

    # Exit a parse tree produced by jacParser#report_action.
    def exitReport_action(self, ctx:jacParser.Report_actionContext):
        pass


    # Enter a parse tree produced by jacParser#walker_action.
    def enterWalker_action(self, ctx:jacParser.Walker_actionContext):
        pass

    # Exit a parse tree produced by jacParser#walker_action.
    def exitWalker_action(self, ctx:jacParser.Walker_actionContext):
        pass


    # Enter a parse tree produced by jacParser#ignore_action.
    def enterIgnore_action(self, ctx:jacParser.Ignore_actionContext):
        pass

    # Exit a parse tree produced by jacParser#ignore_action.
    def exitIgnore_action(self, ctx:jacParser.Ignore_actionContext):
        pass


    # Enter a parse tree produced by jacParser#take_action.
    def enterTake_action(self, ctx:jacParser.Take_actionContext):
        pass

    # Exit a parse tree produced by jacParser#take_action.
    def exitTake_action(self, ctx:jacParser.Take_actionContext):
        pass


    # Enter a parse tree produced by jacParser#expression.
    def enterExpression(self, ctx:jacParser.ExpressionContext):
        pass

    # Exit a parse tree produced by jacParser#expression.
    def exitExpression(self, ctx:jacParser.ExpressionContext):
        pass


    # Enter a parse tree produced by jacParser#assignment.
    def enterAssignment(self, ctx:jacParser.AssignmentContext):
        pass

    # Exit a parse tree produced by jacParser#assignment.
    def exitAssignment(self, ctx:jacParser.AssignmentContext):
        pass


    # Enter a parse tree produced by jacParser#copy_assign.
    def enterCopy_assign(self, ctx:jacParser.Copy_assignContext):
        pass

    # Exit a parse tree produced by jacParser#copy_assign.
    def exitCopy_assign(self, ctx:jacParser.Copy_assignContext):
        pass


    # Enter a parse tree produced by jacParser#inc_assign.
    def enterInc_assign(self, ctx:jacParser.Inc_assignContext):
        pass

    # Exit a parse tree produced by jacParser#inc_assign.
    def exitInc_assign(self, ctx:jacParser.Inc_assignContext):
        pass


    # Enter a parse tree produced by jacParser#connect.
    def enterConnect(self, ctx:jacParser.ConnectContext):
        pass

    # Exit a parse tree produced by jacParser#connect.
    def exitConnect(self, ctx:jacParser.ConnectContext):
        pass


    # Enter a parse tree produced by jacParser#logical.
    def enterLogical(self, ctx:jacParser.LogicalContext):
        pass

    # Exit a parse tree produced by jacParser#logical.
    def exitLogical(self, ctx:jacParser.LogicalContext):
        pass


    # Enter a parse tree produced by jacParser#compare.
    def enterCompare(self, ctx:jacParser.CompareContext):
        pass

    # Exit a parse tree produced by jacParser#compare.
    def exitCompare(self, ctx:jacParser.CompareContext):
        pass


    # Enter a parse tree produced by jacParser#cmp_op.
    def enterCmp_op(self, ctx:jacParser.Cmp_opContext):
        pass

    # Exit a parse tree produced by jacParser#cmp_op.
    def exitCmp_op(self, ctx:jacParser.Cmp_opContext):
        pass


    # Enter a parse tree produced by jacParser#nin.
    def enterNin(self, ctx:jacParser.NinContext):
        pass

    # Exit a parse tree produced by jacParser#nin.
    def exitNin(self, ctx:jacParser.NinContext):
        pass


    # Enter a parse tree produced by jacParser#arithmetic.
    def enterArithmetic(self, ctx:jacParser.ArithmeticContext):
        pass

    # Exit a parse tree produced by jacParser#arithmetic.
    def exitArithmetic(self, ctx:jacParser.ArithmeticContext):
        pass


    # Enter a parse tree produced by jacParser#term.
    def enterTerm(self, ctx:jacParser.TermContext):
        pass

    # Exit a parse tree produced by jacParser#term.
    def exitTerm(self, ctx:jacParser.TermContext):
        pass


    # Enter a parse tree produced by jacParser#factor.
    def enterFactor(self, ctx:jacParser.FactorContext):
        pass

    # Exit a parse tree produced by jacParser#factor.
    def exitFactor(self, ctx:jacParser.FactorContext):
        pass


    # Enter a parse tree produced by jacParser#power.
    def enterPower(self, ctx:jacParser.PowerContext):
        pass

    # Exit a parse tree produced by jacParser#power.
    def exitPower(self, ctx:jacParser.PowerContext):
        pass


    # Enter a parse tree produced by jacParser#atom.
    def enterAtom(self, ctx:jacParser.AtomContext):
        pass

    # Exit a parse tree produced by jacParser#atom.
    def exitAtom(self, ctx:jacParser.AtomContext):
        pass


    # Enter a parse tree produced by jacParser#atom_trailer.
    def enterAtom_trailer(self, ctx:jacParser.Atom_trailerContext):
        pass

    # Exit a parse tree produced by jacParser#atom_trailer.
    def exitAtom_trailer(self, ctx:jacParser.Atom_trailerContext):
        pass


    # Enter a parse tree produced by jacParser#ref.
    def enterRef(self, ctx:jacParser.RefContext):
        pass

    # Exit a parse tree produced by jacParser#ref.
    def exitRef(self, ctx:jacParser.RefContext):
        pass


    # Enter a parse tree produced by jacParser#deref.
    def enterDeref(self, ctx:jacParser.DerefContext):
        pass

    # Exit a parse tree produced by jacParser#deref.
    def exitDeref(self, ctx:jacParser.DerefContext):
        pass


    # Enter a parse tree produced by jacParser#built_in.
    def enterBuilt_in(self, ctx:jacParser.Built_inContext):
        pass

    # Exit a parse tree produced by jacParser#built_in.
    def exitBuilt_in(self, ctx:jacParser.Built_inContext):
        pass


    # Enter a parse tree produced by jacParser#cast_built_in.
    def enterCast_built_in(self, ctx:jacParser.Cast_built_inContext):
        pass

    # Exit a parse tree produced by jacParser#cast_built_in.
    def exitCast_built_in(self, ctx:jacParser.Cast_built_inContext):
        pass


    # Enter a parse tree produced by jacParser#obj_built_in.
    def enterObj_built_in(self, ctx:jacParser.Obj_built_inContext):
        pass

    # Exit a parse tree produced by jacParser#obj_built_in.
    def exitObj_built_in(self, ctx:jacParser.Obj_built_inContext):
        pass


    # Enter a parse tree produced by jacParser#dict_built_in.
    def enterDict_built_in(self, ctx:jacParser.Dict_built_inContext):
        pass

    # Exit a parse tree produced by jacParser#dict_built_in.
    def exitDict_built_in(self, ctx:jacParser.Dict_built_inContext):
        pass


    # Enter a parse tree produced by jacParser#list_built_in.
    def enterList_built_in(self, ctx:jacParser.List_built_inContext):
        pass

    # Exit a parse tree produced by jacParser#list_built_in.
    def exitList_built_in(self, ctx:jacParser.List_built_inContext):
        pass


    # Enter a parse tree produced by jacParser#string_built_in.
    def enterString_built_in(self, ctx:jacParser.String_built_inContext):
        pass

    # Exit a parse tree produced by jacParser#string_built_in.
    def exitString_built_in(self, ctx:jacParser.String_built_inContext):
        pass


    # Enter a parse tree produced by jacParser#node_edge_ref.
    def enterNode_edge_ref(self, ctx:jacParser.Node_edge_refContext):
        pass

    # Exit a parse tree produced by jacParser#node_edge_ref.
    def exitNode_edge_ref(self, ctx:jacParser.Node_edge_refContext):
        pass


    # Enter a parse tree produced by jacParser#node_ref.
    def enterNode_ref(self, ctx:jacParser.Node_refContext):
        pass

    # Exit a parse tree produced by jacParser#node_ref.
    def exitNode_ref(self, ctx:jacParser.Node_refContext):
        pass


    # Enter a parse tree produced by jacParser#walker_ref.
    def enterWalker_ref(self, ctx:jacParser.Walker_refContext):
        pass

    # Exit a parse tree produced by jacParser#walker_ref.
    def exitWalker_ref(self, ctx:jacParser.Walker_refContext):
        pass


    # Enter a parse tree produced by jacParser#graph_ref.
    def enterGraph_ref(self, ctx:jacParser.Graph_refContext):
        pass

    # Exit a parse tree produced by jacParser#graph_ref.
    def exitGraph_ref(self, ctx:jacParser.Graph_refContext):
        pass


    # Enter a parse tree produced by jacParser#edge_ref.
    def enterEdge_ref(self, ctx:jacParser.Edge_refContext):
        pass

    # Exit a parse tree produced by jacParser#edge_ref.
    def exitEdge_ref(self, ctx:jacParser.Edge_refContext):
        pass


    # Enter a parse tree produced by jacParser#edge_to.
    def enterEdge_to(self, ctx:jacParser.Edge_toContext):
        pass

    # Exit a parse tree produced by jacParser#edge_to.
    def exitEdge_to(self, ctx:jacParser.Edge_toContext):
        pass


    # Enter a parse tree produced by jacParser#edge_from.
    def enterEdge_from(self, ctx:jacParser.Edge_fromContext):
        pass

    # Exit a parse tree produced by jacParser#edge_from.
    def exitEdge_from(self, ctx:jacParser.Edge_fromContext):
        pass


    # Enter a parse tree produced by jacParser#edge_any.
    def enterEdge_any(self, ctx:jacParser.Edge_anyContext):
        pass

    # Exit a parse tree produced by jacParser#edge_any.
    def exitEdge_any(self, ctx:jacParser.Edge_anyContext):
        pass


    # Enter a parse tree produced by jacParser#list_val.
    def enterList_val(self, ctx:jacParser.List_valContext):
        pass

    # Exit a parse tree produced by jacParser#list_val.
    def exitList_val(self, ctx:jacParser.List_valContext):
        pass


    # Enter a parse tree produced by jacParser#index_slice.
    def enterIndex_slice(self, ctx:jacParser.Index_sliceContext):
        pass

    # Exit a parse tree produced by jacParser#index_slice.
    def exitIndex_slice(self, ctx:jacParser.Index_sliceContext):
        pass


    # Enter a parse tree produced by jacParser#dict_val.
    def enterDict_val(self, ctx:jacParser.Dict_valContext):
        pass

    # Exit a parse tree produced by jacParser#dict_val.
    def exitDict_val(self, ctx:jacParser.Dict_valContext):
        pass


    # Enter a parse tree produced by jacParser#kv_pair.
    def enterKv_pair(self, ctx:jacParser.Kv_pairContext):
        pass

    # Exit a parse tree produced by jacParser#kv_pair.
    def exitKv_pair(self, ctx:jacParser.Kv_pairContext):
        pass


    # Enter a parse tree produced by jacParser#spawn.
    def enterSpawn(self, ctx:jacParser.SpawnContext):
        pass

    # Exit a parse tree produced by jacParser#spawn.
    def exitSpawn(self, ctx:jacParser.SpawnContext):
        pass


    # Enter a parse tree produced by jacParser#spawn_object.
    def enterSpawn_object(self, ctx:jacParser.Spawn_objectContext):
        pass

    # Exit a parse tree produced by jacParser#spawn_object.
    def exitSpawn_object(self, ctx:jacParser.Spawn_objectContext):
        pass


    # Enter a parse tree produced by jacParser#node_spawn.
    def enterNode_spawn(self, ctx:jacParser.Node_spawnContext):
        pass

    # Exit a parse tree produced by jacParser#node_spawn.
    def exitNode_spawn(self, ctx:jacParser.Node_spawnContext):
        pass


    # Enter a parse tree produced by jacParser#graph_spawn.
    def enterGraph_spawn(self, ctx:jacParser.Graph_spawnContext):
        pass

    # Exit a parse tree produced by jacParser#graph_spawn.
    def exitGraph_spawn(self, ctx:jacParser.Graph_spawnContext):
        pass


    # Enter a parse tree produced by jacParser#walker_spawn.
    def enterWalker_spawn(self, ctx:jacParser.Walker_spawnContext):
        pass

    # Exit a parse tree produced by jacParser#walker_spawn.
    def exitWalker_spawn(self, ctx:jacParser.Walker_spawnContext):
        pass


    # Enter a parse tree produced by jacParser#spawn_ctx.
    def enterSpawn_ctx(self, ctx:jacParser.Spawn_ctxContext):
        pass

    # Exit a parse tree produced by jacParser#spawn_ctx.
    def exitSpawn_ctx(self, ctx:jacParser.Spawn_ctxContext):
        pass


    # Enter a parse tree produced by jacParser#filter_ctx.
    def enterFilter_ctx(self, ctx:jacParser.Filter_ctxContext):
        pass

    # Exit a parse tree produced by jacParser#filter_ctx.
    def exitFilter_ctx(self, ctx:jacParser.Filter_ctxContext):
        pass


    # Enter a parse tree produced by jacParser#spawn_assign.
    def enterSpawn_assign(self, ctx:jacParser.Spawn_assignContext):
        pass

    # Exit a parse tree produced by jacParser#spawn_assign.
    def exitSpawn_assign(self, ctx:jacParser.Spawn_assignContext):
        pass


    # Enter a parse tree produced by jacParser#filter_compare.
    def enterFilter_compare(self, ctx:jacParser.Filter_compareContext):
        pass

    # Exit a parse tree produced by jacParser#filter_compare.
    def exitFilter_compare(self, ctx:jacParser.Filter_compareContext):
        pass


    # Enter a parse tree produced by jacParser#any_type.
    def enterAny_type(self, ctx:jacParser.Any_typeContext):
        pass

    # Exit a parse tree produced by jacParser#any_type.
    def exitAny_type(self, ctx:jacParser.Any_typeContext):
        pass


    # Enter a parse tree produced by jacParser#dot_graph.
    def enterDot_graph(self, ctx:jacParser.Dot_graphContext):
        pass

    # Exit a parse tree produced by jacParser#dot_graph.
    def exitDot_graph(self, ctx:jacParser.Dot_graphContext):
        pass


    # Enter a parse tree produced by jacParser#dot_stmt_list.
    def enterDot_stmt_list(self, ctx:jacParser.Dot_stmt_listContext):
        pass

    # Exit a parse tree produced by jacParser#dot_stmt_list.
    def exitDot_stmt_list(self, ctx:jacParser.Dot_stmt_listContext):
        pass


    # Enter a parse tree produced by jacParser#dot_stmt.
    def enterDot_stmt(self, ctx:jacParser.Dot_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#dot_stmt.
    def exitDot_stmt(self, ctx:jacParser.Dot_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#dot_attr_stmt.
    def enterDot_attr_stmt(self, ctx:jacParser.Dot_attr_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#dot_attr_stmt.
    def exitDot_attr_stmt(self, ctx:jacParser.Dot_attr_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#dot_attr_list.
    def enterDot_attr_list(self, ctx:jacParser.Dot_attr_listContext):
        pass

    # Exit a parse tree produced by jacParser#dot_attr_list.
    def exitDot_attr_list(self, ctx:jacParser.Dot_attr_listContext):
        pass


    # Enter a parse tree produced by jacParser#dot_a_list.
    def enterDot_a_list(self, ctx:jacParser.Dot_a_listContext):
        pass

    # Exit a parse tree produced by jacParser#dot_a_list.
    def exitDot_a_list(self, ctx:jacParser.Dot_a_listContext):
        pass


    # Enter a parse tree produced by jacParser#dot_edge_stmt.
    def enterDot_edge_stmt(self, ctx:jacParser.Dot_edge_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#dot_edge_stmt.
    def exitDot_edge_stmt(self, ctx:jacParser.Dot_edge_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#dot_edgeRHS.
    def enterDot_edgeRHS(self, ctx:jacParser.Dot_edgeRHSContext):
        pass

    # Exit a parse tree produced by jacParser#dot_edgeRHS.
    def exitDot_edgeRHS(self, ctx:jacParser.Dot_edgeRHSContext):
        pass


    # Enter a parse tree produced by jacParser#dot_edgeop.
    def enterDot_edgeop(self, ctx:jacParser.Dot_edgeopContext):
        pass

    # Exit a parse tree produced by jacParser#dot_edgeop.
    def exitDot_edgeop(self, ctx:jacParser.Dot_edgeopContext):
        pass


    # Enter a parse tree produced by jacParser#dot_node_stmt.
    def enterDot_node_stmt(self, ctx:jacParser.Dot_node_stmtContext):
        pass

    # Exit a parse tree produced by jacParser#dot_node_stmt.
    def exitDot_node_stmt(self, ctx:jacParser.Dot_node_stmtContext):
        pass


    # Enter a parse tree produced by jacParser#dot_node_id.
    def enterDot_node_id(self, ctx:jacParser.Dot_node_idContext):
        pass

    # Exit a parse tree produced by jacParser#dot_node_id.
    def exitDot_node_id(self, ctx:jacParser.Dot_node_idContext):
        pass


    # Enter a parse tree produced by jacParser#dot_port.
    def enterDot_port(self, ctx:jacParser.Dot_portContext):
        pass

    # Exit a parse tree produced by jacParser#dot_port.
    def exitDot_port(self, ctx:jacParser.Dot_portContext):
        pass


    # Enter a parse tree produced by jacParser#dot_subgraph.
    def enterDot_subgraph(self, ctx:jacParser.Dot_subgraphContext):
        pass

    # Exit a parse tree produced by jacParser#dot_subgraph.
    def exitDot_subgraph(self, ctx:jacParser.Dot_subgraphContext):
        pass


    # Enter a parse tree produced by jacParser#dot_id.
    def enterDot_id(self, ctx:jacParser.Dot_idContext):
        pass

    # Exit a parse tree produced by jacParser#dot_id.
    def exitDot_id(self, ctx:jacParser.Dot_idContext):
        pass



del jacParser


"""Built in actions for Jaseci"""
import numpy as np
from operator import itemgetter
from jaseci.actions.live_actions import jaseci_action


@jaseci_action()
def cosine_sim(vec_a: list, vec_b: list):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector

    Return - float between 0 and 1
    """
    result = np.dot(vec_a, vec_b) / (np.linalg.norm(vec_a) *
                                     np.linalg.norm(vec_b))
    return result.astype(float)


@jaseci_action()
def dot_product(vec_a: list, vec_b: list):
    """
    Caculate the dot product of two given vectors
    Param 1 - First vector
    Param 2 - Second vector

    Return - float between 0 and 1
    """

    return np.inner(vec_a, vec_b).tolist()


@jaseci_action()
def get_centroid(vec_list: list):
    """
    Calculate the centroid of the given list of vectors
    Param 1 - List of vectors

    Return - (centroid vector, cluster tightness)
    """
    centroid = np.mean(vec_list, axis=0)
    tightness = np.mean([cosine_sim(vec, centroid)
                         for vec in vec_list]).astype(float)
    return [centroid, tightness]


@jaseci_action()
def softmax(vec_list: list):
    """
    Calculate the centroid of the given list of vectors
    Param 1 - List of vectors

    Return - (centroid vector, cluster tightness)
    """
    e_x = np.exp(vec_list - np.max(vec_list))
    return list(e_x / e_x.sum())


@jaseci_action()
def sort_by_key(data: dict, reverse=False, key_pos=None):
    """
    Sort the given list. Optionally by specific key
    Param 1 - List of items
    Param 2 - if Reverse
    Param 3 (Optional) - Index of the key to be used for sorting
    if param 1 is a list of tuples.

    Deprecated
    """
    if (key_pos is not None):
        return sorted(data, key=itemgetter(key_pos), reverse=reverse)
    else:
        return sorted(data, reverse=reverse)



"""
General action base class with automation for hot loading
"""
from importlib.util import spec_from_file_location, module_from_spec
from jaseci.utils.utils import logger
from jaseci.actions.remote_actions import ACTIONS_SPEC_LOC
from jaseci.actions.remote_actions import serv_actions, mark_as_remote
import requests
import os
import sys
import inspect

live_actions = {}


def jaseci_action(act_group=None, aliases=list(), allow_remote=False):
    """Decorator for Jaseci Action interface"""
    caller_globals = dict(inspect.getmembers(
        inspect.stack()[1][0]))["f_globals"]
    if allow_remote and 'serv_actions' not in caller_globals:
        caller_globals['serv_actions'] = serv_actions

    def decorator_func(func):
        if(allow_remote):
            mark_as_remote([func, act_group, aliases, caller_globals])
        return assimilate_action(func, act_group)
    return decorator_func


def assimilate_action(func, act_group=None):
    """Helper for jaseci_action decorator"""
    act_group = [func.__module__.split(
        '.')[-1]] if act_group is None else act_group
    live_actions[f"{'.'.join(act_group+[func.__name__])}"] = func
    return func


def load_local_actions(file):
    """Load all jaseci actions from python file"""
    spec = spec_from_file_location(str(file).split("/")[-1][:-3], str(file))
    if(spec is None):
        logger.error(f"Cannot hot load from action file {file}")
        return False
    else:
        module_dir = os.path.dirname(os.path.realpath(str(file)))
        if module_dir not in sys.path:
            sys.path.append(module_dir)
        spec.loader.exec_module(module_from_spec(spec))
        return True


def load_standard():
    import jaseci.actions.standard.net  # noqa
    import jaseci.actions.standard.rand  # noqa
    import jaseci.actions.standard.request  # noqa
    import jaseci.actions.standard.std  # noqa
    import jaseci.actions.standard.file  # noqa
    import jaseci.actions.standard.vector  # noqa
    import jaseci.actions.standard.date  # noqa


load_standard()


def load_preconfig_actions(hook):
    import json
    action_preload = hook.resolve_glob('ACTION_SETS', None)
    if (action_preload):
        action_preload = json.loads(action_preload)
        for i in action_preload['local']:
            load_local_actions(i)
        for i in action_preload['remote']:
            load_remote_actions(i)


def get_global_actions(hook):
    """
    Loads all global action hooks for use by Jac programs
    Attaches globals to mem_hook
    """
    from jaseci.attr.action import action
    import uuid
    global_action_list = []
    for i in live_actions.keys():
        if(i.startswith('std.') or i.startswith('file.') or
           i.startswith('net.') or i.startswith('rand.') or
           i.startswith('vector.') or i.startswith('request.')):
            global_action_list.append(
                action(
                    m_id=uuid.UUID(int=0).urn,
                    h=hook,
                    mode='public',
                    name=i,
                    value=i,
                    persist=False
                ).jid
            )
    return global_action_list


def load_remote_actions(url):
    """Load all jaseci actions from live pod"""
    headers = {'content-type': 'application/json'}
    try:
        spec = requests.get(url.rstrip('/')+ACTIONS_SPEC_LOC,
                            headers=headers)
        spec = spec.json()
        for i in spec.keys():
            live_actions[i] = gen_remote_func_hook(url, i, spec[i])
        return True

    except Exception as e:
        logger.error(f"Cannot hot load remote actions at {url}: {e}")
        return False


def gen_remote_func_hook(url, act_name, param_names):
    """Generater for function calls for remote action calls"""
    def func(*args):
        params = {}
        for i in range(len(param_names)):
            if(i < len(args)):
                params[param_names[i]] = args[i]
            else:
                params[param_names[i]] = None
        act_url = f"{url.rstrip('/')}/{act_name.split('.')[-1]}"
        res = requests.post(
            act_url, headers={'content-type': 'application/json'}, json=params)
        return res.json()
    return func



"""
Main master handler for each user of Jaseci, serves as main interface between
between user and Jaseci
"""

from jaseci.element.element import element
from jaseci.api.alias_api import alias_api
from jaseci.api.object_api import object_api
from jaseci.api.graph_api import graph_api
from jaseci.api.sentinel_api import sentinel_api
from jaseci.api.walker_api import walker_api
from jaseci.api.architype_api import architype_api
from jaseci.api.config_api import config_api
from jaseci.api.interface import interface
from jaseci.api.master_api import master_api
from jaseci.api.jac_api import jac_api


class master(element, interface, master_api, alias_api,
             graph_api, object_api, sentinel_api, walker_api,
             architype_api, jac_api):
    """Main class for master functions for user"""

    def __init__(self, head_master=None, *args, **kwargs):
        kwargs['m_id'] = None

        element.__init__(self, kind="Jaseci Master", *args, **kwargs)
        master_api.__init__(self, head_master)
        alias_api.__init__(self)
        config_api.__init__(self)
        graph_api.__init__(self)
        walker_api.__init__(self)
        sentinel_api.__init__(self)
        interface.__init__(self)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        graph_api.destroy(self)
        sentinel_api.destroy(self)
        master_api.destroy(self)
        walker_api.destroy(self)
        super().destroy()



"""
Interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.utils.utils import is_jsonable, is_urn, parse_str_token
from jaseci.element.element import element
from jaseci.graph.node import node
from jaseci.graph.edge import edge
from jaseci.attr.action import action
from jaseci.jac.jac_set import jac_set
from jaseci.jac.ir.jac_code import jac_ast_to_ir, jac_ir_to_ast
from jaseci.jac.machine.jac_scope import jac_scope
from jaseci.jac.machine.machine_state import machine_state

from jaseci.jac.machine.jac_value import jac_value
from jaseci.jac.machine.jac_value import jac_elem_unwrap as jeu
from copy import copy


class interp(machine_state):
    """Shared interpreter class across both sentinels and walkers"""

    def run_attr_stmt(self, jac_ast, obj):
        """
        attr_stmt: has_stmt | can_stmt;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'has_stmt'):
            self.run_has_stmt(kid[0], obj)
        elif(kid[0].name == 'can_stmt'):
            self.run_can_stmt(kid[0], obj)

    def run_has_stmt(self, jac_ast, obj):
        """
        has_stmt:
                KW_HAS KW_PRIVATE? KW_ANCHOR? has_assign
                (COMMA has_assign)* SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        kid = kid[1:]
        is_private = False
        is_anchor = False
        while True:
            if(kid[0].name == 'KW_PRIVATE'):
                kid = kid[1:]
                is_private = True
            if(kid[0].name == 'KW_ANCHOR'):
                kid = kid[1:]
                is_anchor = True
            self.run_has_assign(kid[0], obj, is_private, is_anchor)
            kid = kid[1:]
            if(not len(kid) or kid[0].name != 'COMMA'):
                break
            else:
                kid = kid[1:]

    def run_has_assign(self, jac_ast, obj, is_private, is_anchor):
        """
        has_assign: NAME | NAME EQ expression;
        """
        kid = self.set_cur_ast(jac_ast)
        var_name = kid[0].token_text()
        var_val = None  # jac's null
        if(len(kid) > 1):
            var_val = self.run_expression(kid[2]).value
        if(is_anchor):
            if('anchor' in dir(obj)):
                if(obj.anchor is None):
                    obj.anchor = var_name
            else:
                self.rt_error('anchors not allowed for this type',
                              kid[0])

        if(var_name == '_private'):
            self.rt_error(
                f'Has variable name of `_private` not allowed!', kid[0])
        elif (var_name not in obj.context.keys()):  # Runs only once
            jac_value(self, ctx=obj,
                      name=var_name, value=var_val).write(kid[0], force=True)
        if(is_private):
            if('_private' in obj.context.keys()):
                if(var_name not in obj.context['_private']):
                    obj.context['_private'].append(var_name)
            else:
                obj.context['_private'] = [var_name]

    def run_can_stmt(self, jac_ast, obj):
        """
        can_stmt:
            KW_CAN dotted_name (preset_in_out event_clause)? (
                COMMA dotted_name (preset_in_out event_clause)?
            )* SEMI
            | KW_CAN NAME event_clause? code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        kid = kid[1:]
        while True:
            action_type = 'activity'
            access_list = None
            preset_in_out = None
            if (kid[0].name == 'NAME'):
                action_name = kid[0].token_text()
            else:
                action_name = self.run_dotted_name(kid[0])
            kid = kid[1:]
            if(len(kid) > 0 and kid[0].name == 'preset_in_out'):
                preset_in_out = jac_ast_to_ir(kid[0])
                kid = kid[1:]
            if(len(kid) > 0 and kid[0].name == 'event_clause'):
                action_type, access_list = self.run_event_clause(kid[0])
                kid = kid[1:]
            if (not isinstance(obj, node) and action_type != 'activity'):
                self.rt_warn(
                    "Only nodes can have on entry/exit, treating as activity",
                    kid[0])
                action_type = 'activity'
            if (kid[0].name == 'code_block'):
                act = action(
                    m_id=self._m_id,
                    h=self._h,
                    name=action_name,
                    value=jac_ast_to_ir(kid[0]),
                    preset_in_out=preset_in_out,
                    access_list=access_list
                )
                getattr(obj, f"{action_type}_action_ids").add_obj(act)
                self._jac_scope.add_action(act)
                break
            else:
                func_link = \
                    self.get_builtin_action(action_name, jac_ast)
                if(func_link):
                    act = action(
                        m_id=self._m_id,
                        h=self._h,
                        name=action_name,
                        value=func_link,
                        preset_in_out=preset_in_out,
                        access_list=access_list
                    )
                    getattr(obj, f"{action_type}_action_ids").add_obj(act)
                    self._jac_scope.add_action(act)
            if(not len(kid) or kid[0].name != 'COMMA'):
                break
            else:
                kid = kid[1:]

    def run_event_clause(self, jac_ast):
        """
        event_clause:
                KW_WITH name_list? (KW_ENTRY | KW_EXIT | KW_ACTIVITY);
        """
        kid = self.set_cur_ast(jac_ast)
        nl = []
        if(kid[1].name == "name_list"):
            nl = self.run_name_list(kid[1])
        return kid[-1].token_text(), nl

    def run_dotted_name(self, jac_ast):
        """
        dotted_name: NAME (DOT NAME)*;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = ''
        for i in kid:
            if(i.name == 'NAME'):
                ret += i.token_text()
                if(i == kid[-1]):
                    break
                ret += '.'
        return ret

    def run_name_list(self, jac_ast):
        """
        name_list: NAME (COMMA NAME)*;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = []
        for i in kid:
            if(i.name == 'NAME'):
                ret.append(i.token_text())
        return ret

    def run_expr_list(self, jac_ast, wrap=False):
        """
        expr_list: expression (COMMA expression)*;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = []
        for i in kid:
            if(i.name == 'expression'):
                if(wrap):
                    ret.append(self.run_expression(i).wrap())
                else:
                    ret.append(self.run_expression(i).value)
        return jac_value(self, value=ret)

    def run_code_block(self, jac_ast):
        """
        code_block: LBRACE statement* RBRACE | COLON statement;
        TODO: Handle breaks and continues
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if (self._loop_ctrl):
                if (self._loop_ctrl == 'continue'):
                    self._loop_ctrl = None
                return
            if(i.name == 'statement'):
                self.run_statement(jac_ast=i)

    def run_statement(self, jac_ast):
        """
        statement:
            code_block
            | node_ctx_block
            | expression SEMI
            | if_stmt
            | try_stmt
            | for_stmt
            | while_stmt
            | assert_stmt SEMI
            | ctrl_stmt SEMI
            | destroy_action
            | report_action
            | walker_action;
        """
        if (self._stopped):
            return
        kid = self.set_cur_ast(jac_ast)
        self.run_rule(kid[0])

    def run_if_stmt(self, jac_ast):
        """
        if_stmt: KW_IF expression code_block elif_stmt* else_stmt?;
        """
        kid = self.set_cur_ast(jac_ast)
        if(self.run_expression(kid[1]).value):
            self.run_code_block(kid[2])
            return
        kid = kid[3:]
        if(len(kid)):
            while True:
                if(kid[0].name == 'elif_stmt'):
                    if(self.run_elif_stmt(kid[0])):
                        return
                elif(kid[0].name == 'else_stmt'):
                    self.run_else_stmt(kid[0])
                    return
                kid = kid[1:]
                if(not len(kid)):
                    break

    def run_try_stmt(self, jac_ast):
        """
        try_stmt: KW_TRY code_block else_from_try?;
        """
        kid = self.set_cur_ast(jac_ast)
        try:
            self.run_code_block(kid[1])
            return
        except Exception as e:
            if(len(kid) > 2):
                self.run_else_from_try(kid[2], e)

    def run_else_from_try(self, jac_ast, e):
        """
        else_from_try:
            KW_ELSE (LPAREN NAME RPAREN)? code_block
            | KW_ELSE (KW_WITH NAME)? code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        if(len(kid) > 2):
            jac_value(self, ctx=self._jac_scope.local_scope,
                      name=kid[2].token_text(),
                      value=self.jac_exception(e, jac_ast)).write(kid[2])
        self.run_code_block(kid[-1])

    def run_elif_stmt(self, jac_ast):
        """
        elif_stmt: KW_ELIF expression code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        if(self.run_expression(kid[1]).value):
            self.run_code_block(kid[2])
            return True
        else:
            return False

    def run_else_stmt(self, jac_ast):
        """
        else_stmt: KW_ELSE code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        self.run_code_block(kid[1])

    def run_for_stmt(self, jac_ast):
        """
        for_stmt:
            KW_FOR expression KW_TO expression KW_BY expression code_block
            | KW_FOR NAME KW_IN expression code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        loops = 0
        if(kid[1].name == 'expression'):
            self.run_expression(kid[1])
            while self.run_expression(kid[3]).value:
                self.run_code_block(kid[6])
                loops += 1
                if (self._loop_ctrl == 'break'):
                    self._loop_ctrl = None
                    break
                self.run_expression(kid[5])
                if(loops > self._loop_limit):
                    self.rt_error(f'Hit loop limit, breaking...', kid[0])
                    self._loop_ctrl = 'break'
        else:
            var = self._jac_scope.get_live_var(
                kid[1].token_text(), create_mode=True)
            lst = self.run_expression(kid[3]).value
            # should check that lst is list here
            if(not isinstance(lst, list)):
                self.rt_error('Not a list for iteration!', kid[3])
            for i in lst:
                var.value = i
                var.write(kid[1])
                self.run_code_block(kid[4])
                loops += 1
                if (self._loop_ctrl == 'break'):
                    self._loop_ctrl = None
                    break
                if(loops > self._loop_limit):
                    self.rt_error(f'Hit loop limit, breaking...', kid[0])
                    self._loop_ctrl = 'break'

    def run_while_stmt(self, jac_ast):
        """
        while_stmt: KW_WHILE expression code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        loops = 0
        while self.run_expression(kid[1]).value:
            self.run_code_block(kid[2])
            loops += 1
            if (self._loop_ctrl == 'break'):
                self._loop_ctrl = None
                break
            if(loops > self._loop_limit):
                self.rt_error(f'Hit loop limit, breaking...', kid[0])
                self._loop_ctrl = 'break'

    def run_ctrl_stmt(self, jac_ast):
        """
        ctrl_stmt: KW_CONTINUE | KW_BREAK | KW_SKIP;
        """
        kid = self.set_cur_ast(jac_ast)
        if (kid[0].name == 'KW_SKIP'):
            self._stopped = 'skip'
        elif (kid[0].name == 'KW_BREAK'):
            self._loop_ctrl = 'break'
        elif (kid[0].name == 'KW_CONTINUE'):
            self._loop_ctrl = 'continue'

    def run_assert_stmt(self, jac_ast):
        """
        assert_stmt: KW_ASSERT expression;
        """
        kid = self.set_cur_ast(jac_ast)
        passed = False
        try:
            passed = self.run_expression(kid[1]).value
        except Exception as e:
            e = self.jac_exception(e, jac_ast)
            raise Exception('Jac Assert Failed', kid[1].get_text(), e)
        if(not passed):
            raise Exception('Jac Assert Failed', kid[1].get_text())

    def run_destroy_action(self, jac_ast):
        """
        destroy_action: KW_DESTROY expression SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1])
        if (isinstance(result.value, element)):
            self.destroy_node_ids.add_obj(result.value)
        elif (isinstance(result.value, jac_set)):
            self.destroy_node_ids.add_obj_list(result.value)
        result.self_destruct(kid[1])

    def run_report_action(self, jac_ast):
        """
        report_action: KW_REPORT expression SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        report = self.run_expression(kid[1]).wrap(serialize_mode=True)
        if(not is_jsonable(report)):
            self.rt_error(f'Report not Json serializable', kid[0])
        self.report.append(copy(report))

    def run_expression(self, jac_ast):
        """
        expression: connect (assignment | copy_assign | inc_assign)?;
        """
        def check_can_write(val):
            if(val.ctx is None):
                self.rt_error("Cannot assign to this experssion", kid[0])
                return False
            return True
        kid = self.set_cur_ast(jac_ast)
        if(len(kid) == 1):
            return self.run_connect(kid[0])
        else:
            if(kid[1].name == "assignment"):
                self._assign_mode = True
                dest = self.run_connect(kid[0])
                self._assign_mode = False
                if(not check_can_write(dest)):
                    return dest
                return self.run_assignment(kid[1], dest=dest)
            elif(kid[1].name == "copy_assign"):
                dest = self.run_connect(kid[0])
                if(not check_can_write(dest)):
                    return dest
                return self.run_copy_assign(kid[1], dest=dest)
            elif(kid[1].name == "inc_assign"):
                dest = self.run_connect(kid[0])
                if(not check_can_write(dest)):
                    return dest
                return self.run_inc_assign(kid[1], dest=dest)

    def run_assignment(self, jac_ast, dest):
        """
        assignment: EQ expression;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1])
        dest.value = result.value
        dest.write(jac_ast)
        return dest

    def run_copy_assign(self, jac_ast, dest):
        """
        copy_assign: CPY_EQ expression;
        """
        kid = self.set_cur_ast(jac_ast)
        src = self.run_expression(kid[1])
        if (not self.rt_check_type(dest.value, [node, edge], kid[1])):
            self.rt_error("':=' only applies to nodes and edges", kid[1])
            return dest
        if (dest.value.name != src.value.name):
            self.rt_error(
                f"Node/edge arch {dest.value} don't "
                f"match {src.value}!", kid[1])
            return dest
        for i in src.value.context.keys():
            if(i in dest.value.context.keys()):
                jac_value(self, ctx=dest.value, name=i,
                          value=src.value.context[i]).write(jac_ast)
        return dest

    def run_inc_assign(self, jac_ast, dest):
        """
        inc_assign: (PEQ | MEQ | TEQ | DEQ) expression;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'PEQ'):
            dest.value = dest.value + self.run_expression(kid[1]).value
        elif(kid[0].name == 'MEQ'):
            dest.value = dest.value - self.run_expression(kid[1]).value
        elif(kid[0].name == 'TEQ'):
            dest.value = dest.value * self.run_expression(kid[1]).value
        elif(kid[0].name == 'DEQ'):
            dest.value = dest.value / self.run_expression(kid[1]).value
        dest.write(jac_ast)
        return dest

    def run_connect(self, jac_ast):
        """
        connect: logical ( (NOT)? edge_ref expression)?;
        """
        kid = self.set_cur_ast(jac_ast)
        if (len(kid) < 2):
            return self.run_logical(kid[0])
        bret = self.run_logical(kid[0])
        base = bret.value
        tret = self.run_expression(kid[-1])
        target = tret.value
        self.rt_check_type(base, [node, jac_set], kid[0])
        self.rt_check_type(target, [node, jac_set], kid[-1])
        if(isinstance(base, node)):
            base = jac_set(in_list=[base])
        if(isinstance(target, node)):
            target = jac_set(in_list=[target])
        if (kid[1].name == 'NOT'):
            for i in target.obj_list():
                for j in base.obj_list():
                    j.detach_edges(i, self.run_edge_ref(kid[2]).obj_list())
            return bret
        else:
            direction = kid[1].kid[0].name
            for i in target.obj_list():
                for j in base.obj_list():
                    use_edge = self.run_edge_ref(kid[1], is_spawn=True)
                    if (direction == 'edge_from'):
                        j.attach_inbound(i, [use_edge])
                    elif (direction == 'edge_to'):
                        j.attach_outbound(i, [use_edge])
                    else:
                        j.attach_bidirected(i, [use_edge])
        return tret

    def run_logical(self, jac_ast):
        """
        logical: compare ((KW_AND | KW_OR) compare)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_compare(kid[0])
        kid = kid[1:]
        while (kid):
            if (kid[0].name == 'KW_AND'):
                if (result):
                    result.value = result.value and self.run_compare(
                        kid[1]).value
            elif (kid[0].name == 'KW_OR'):
                if (not result):
                    result.value = result.value or self.run_compare(
                        kid[1]).value
            kid = kid[2:]
            if(not kid):
                break
        return result

    def run_compare(self, jac_ast):
        """
        compare: NOT compare | arithmetic (cmp_op arithmetic)*;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'NOT'):
            return jac_value(self, value=not self.run_compare(kid[1]).value)
        else:
            result = self.run_arithmetic(kid[0])
            kid = kid[1:]
            while (kid):
                other_res = self.run_arithmetic(kid[1])
                result = self.run_cmp_op(
                    kid[0], result, other_res)
                kid = kid[2:]
                if(not kid):
                    break
            return result

    def run_cmp_op(self, jac_ast, val1, val2):
        """
        cmp_op: EE | LT | GT | LTE | GTE | NE | KW_IN | nin;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'EE'):
            return jac_value(self, value=val1.value == val2.value)
        elif(kid[0].name == 'LT'):
            return jac_value(self, value=val1.value < val2.value)
        elif(kid[0].name == 'GT'):
            return jac_value(self, value=val1.value > val2.value)
        elif(kid[0].name == 'LTE'):
            return jac_value(self, value=val1.value <= val2.value)
        elif(kid[0].name == 'GTE'):
            return jac_value(self, value=val1.value >= val2.value)
        elif(kid[0].name == 'NE'):
            return jac_value(self, value=val1.value != val2.value)
        elif(kid[0].name == 'KW_IN'):
            return jac_value(self, value=val1.value in val2.value)
        elif(kid[0].name == 'nin'):
            return jac_value(self, value=val1.value not in val2.value)

    def run_arithmetic(self, jac_ast):
        """
        arithmetic: term ((PLUS | MINUS) term)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_term(kid[0])
        kid = kid[1:]
        while (kid):
            other_res = self.run_term(kid[1])
            if(kid[0].name == 'PLUS'):
                result.value = result.value + other_res.value
            elif(kid[0].name == 'MINUS'):
                result.value = result.value - other_res.value
            kid = kid[2:]
            if(not kid):
                break
        return result

    def run_term(self, jac_ast):
        """
        term: factor ((STAR_MUL | DIV | MOD) factor)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_factor(kid[0])
        kid = kid[1:]
        while (kid):
            other_res = self.run_factor(kid[1])
            if(kid[0].name == 'STAR_MUL'):
                result.value = result.value * other_res.value
            elif(kid[0].name == 'DIV'):
                result.value = result.value / other_res.value
            elif(kid[0].name == 'MOD'):
                result.value = result.value % other_res.value
            kid = kid[2:]
            if(not kid):
                break
        return result

    def run_factor(self, jac_ast):
        """
        factor: (PLUS | MINUS) factor | power;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'power'):
            return self.run_power(kid[0])
        else:
            result = self.run_factor(kid[1])
            if(kid[0].name == 'MINUS'):
                result.value = -(result.value)
            return result

    def run_power(self, jac_ast):
        """
        power: atom (POW factor)*;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_atom(kid[0])
        kid = kid[1:]
        if(len(kid) < 1):
            return result
        elif(kid[0].name == 'POW'):
            while (kid):
                result.value = result.value ** self.run_factor(kid[1]).value
                kid = kid[2:]
                if(not kid):
                    break
            return result

    def run_atom(self, jac_ast):
        """
        atom:
            INT
            | FLOAT
            | STRING
            | BOOL
            | NULL
            | NAME
            | node_edge_ref
            | list_val
            | dict_val
            | LPAREN expression RPAREN
            | DBL_COLON NAME spawn_ctx?
            | atom atom_trailer+
            | spawn
            | ref
            | deref
            | any_type;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'INT'):
            return jac_value(self, value=int(kid[0].token_text()))
        elif(kid[0].name == 'FLOAT'):
            return jac_value(self, value=float(kid[0].token_text()))
        elif(kid[0].name == 'STRING'):
            return jac_value(
                self, value=parse_str_token(kid[0].token_text()))
        elif(kid[0].name == 'BOOL'):
            return jac_value(self, value=bool(kid[0].token_text() == 'true'))
        elif(kid[0].name == 'NULL'):
            return jac_value(self, value=None)
        elif(kid[0].name == 'NAME'):
            name = kid[0].token_text()
            val = self._jac_scope.get_live_var(
                name, create_mode=self._assign_mode)
            if(val is None):
                self.rt_error(f"Variable not defined - {name}", kid[0])
                return jac_value(self, )
            return val
        elif(kid[0].name == 'LPAREN'):
            return self.run_expression(kid[1])
        elif(kid[0].name == 'DBL_COLON'):
            return self.run_atom_trailer(jac_ast, None)
        elif(kid[0].name == 'atom'):
            ret = self.run_atom(kid[0])
            for i in kid[1:]:
                ret = self.run_atom_trailer(i, ret)
            return ret
        else:
            return self.run_rule(kid[0])

    def run_atom_trailer(self, jac_ast, atom_res):
        """
        atom_trailer:
            DOT built_in
            | DOT NAME
            | index_slice
            | LPAREN expr_list? RPAREN
            | DBL_COLON NAME spawn_ctx?;
        """
        kid = self.set_cur_ast(jac_ast)
        if(atom_res is None):
            atom_res = jac_value(self, value=self._jac_scope.has_obj)
        if(kid[0].name == 'DOT'):
            if(kid[1].name == 'built_in'):

                return self.run_built_in(kid[1], atom_res)
            elif(kid[1].name == 'NAME'):
                d = atom_res.value
                n = kid[1].token_text()
                if(self.rt_check_type(d, [dict, element], kid[0])):
                    ret = jac_value(self, ctx=d, name=n)
                    ret.unwrap()
                    return ret
                else:
                    self.rt_error(f"Invalid variable {n}", kid[0])
        elif (kid[0].name == "index_slice"):
            if(not self.rt_check_type(
                    atom_res.value, [list, str, dict], kid[0])):
                return atom_res
            return self.run_index_slice(kid[0], atom_res)
        elif(kid[0].name == "LPAREN"):
            param_list = []
            if(kid[1].name == 'expr_list'):
                param_list = self.run_expr_list(kid[1]).value
            if (isinstance(atom_res.value, action)):
                try:
                    ret = atom_res.value.trigger(
                        param_list, self._jac_scope, self)
                except Exception as e:
                    self.rt_error(f'{e}', jac_ast)
                    ret = None
                return jac_value(self, value=ret)
            else:
                self.rt_error(f'Unable to execute ability {atom_res}',
                              kid[0])
        elif (kid[0].name == 'DBL_COLON'):
            if(len(kid) > 2):
                self.run_spawn_ctx(kid[2], atom_res.value)
            self.call_ability(
                nd=atom_res.value,
                name=kid[1].token_text(),
                act_list=atom_res.value.activity_action_ids)
            return atom_res

    def run_ref(self, jac_ast):
        """
        ref: '&' expression;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1])
        if (self.rt_check_type(result.value, element, kid[1])):
            result = jac_value(self, value=result.value.jid)
        return result

    def run_deref(self, jac_ast):
        """
        deref: '*' expression;
        """
        kid = self.set_cur_ast(jac_ast)
        result = self.run_expression(kid[1])
        if (is_urn(result.value)):
            result = jac_value(
                self, value=jeu(result.value.replace('urn', 'jac'), self))
        else:
            self.rt_error(f'{result.value} not valid reference', kid[1])
        return result

    def run_built_in(self, jac_ast, atom_res):
        """
        built_in:
            cast_built_in
            | obj_built_in
            | dict_built_in
            | list_built_in
            | string_built_in;
        """
        return self.run_rule(jac_ast.kid[0], atom_res)

    def run_cast_built_in(self, jac_ast, atom_res):
        """
        arch_built_in: any_type;
        """
        kid = self.set_cur_ast(jac_ast)
        typ = self.run_any_type(kid[0])
        if (typ.value == edge):
            if(isinstance(atom_res.value, node)):
                return jac_value(self, value=self.obj_set_to_jac_set(
                    self.current_node.attached_edges(atom_res.value)))
            elif(isinstance(atom_res.value, edge)):
                return atom_res
            elif(isinstance(atom_res.value, jac_set)):
                res = jac_set()
                for i in atom_res.value.obj_list():
                    if(isinstance(i, edge)):
                        res.add_obj(i)
                    elif(isinstance(i, node)):
                        res += self.obj_set_to_jac_set(
                            self.current_node.attached_edges(i))
                return jac_value(self, value=res)
            else:
                self.rt_error(f'Cannot get edges from {atom_res.value}. '
                              f'Type {atom_res.jac_type()} invalid', kid[0])
        # may want to remove 'here" node from return below
        elif (typ.value == node):
            if(isinstance(atom_res.value, node)):
                return atom_res
            elif(isinstance(atom_res.value, edge)):
                return jac_value(self, value=self.obj_set_to_jac_set(
                    atom_res.nodes()))
            elif(isinstance(atom_res.value, jac_set)):
                res = jac_set()
                for i in atom_res.value.obj_list():
                    if(isinstance(i, edge)):
                        res.add_obj(i.to_node())
                        res.add_obj(i.from_node())
                    elif(isinstance(i, node)):
                        res.add_obj(i)
                return jac_value(self, value=res)
            else:
                self.rt_error(f'Cannot get nodes from {atom_res}. '
                              f'Type {atom_res.jac_type()} invalid', kid[0])
        else:
            try:
                atom_res.value = typ.value(atom_res.value)
            except Exception:
                self.rt_error(
                    f'Invalid cast of {atom_res.jac_type()} '
                    f'to {typ.wrap()}', kid[0])
            return atom_res

        return atom_res

    def run_obj_built_in(self, jac_ast, atom_res):
        """
        obj_built_in: KW_CONTEXT | KW_INFO | KW_DETAILS;
        """
        kid = self.set_cur_ast(jac_ast)
        from jaseci.actor.walker import walker
        if (kid[0].name == "KW_CONTEXT"):
            if(self.rt_check_type(atom_res.value,
                                  [node, edge, walker], kid[0])):
                return jac_value(self, value=atom_res.value.context)
        elif (kid[0].name == "KW_INFO"):
            if(self.rt_check_type(atom_res.value,
                                  [node, edge, walker], kid[0])):
                return jac_value(
                    self, value=atom_res.value.serialize(detailed=False))
        elif (kid[0].name == "KW_DETAILS"):
            if(self.rt_check_type(atom_res.value,
                                  [node, edge, walker], kid[0])):
                return jac_value(
                    self, value=atom_res.value.serialize(detailed=True))
        return atom_res

    def run_dict_built_in(self, jac_ast, atom_res):
        """
        dict_built_in:
            KW_KEYS
            | LBRACE name_list RBRACE
            | (TYP_DICT DBL_COLON | DICT_DBL_COLON) NAME (
                LPAREN expr_list RPAREN
            )?;
        """
        kid = self.set_cur_ast(jac_ast)
        if (kid[0].name == "KW_KEYS"):
            if(isinstance(atom_res.value, dict)):
                return jac_value(self, value=list(atom_res.value.keys()))
            else:
                self.rt_error(f'Cannot get keys of {atom_res}. '
                              f'Not Dictionary!', kid[0])
                return jac_value(self, value=[])
        elif(len(kid) > 1 and kid[1].name == 'name_list'):
            filter_on = self.run_name_list(kid[1])
            d = atom_res.value
            if(self.rt_check_type(d, [dict], kid[0])):
                d = {k: d[k] for k in d if k in filter_on}
                return jac_value(self, value=d)
        else:
            if(not self.rt_check_type(atom_res.value, [dict], kid[0])):
                return atom_res
            kid = kid[1:]
            if(kid[0].name == "DBL_COLON"):
                kid = kid[1:]
            result = None
            op = kid[0].token_text()
            try:
                if (op == "items"):
                    result = jac_value(self, value=list(
                        map(list, atom_res.value.items())))
                elif (op == "copy"):
                    result = jac_value(self, value=atom_res.value.copy())
                elif (op == "keys"):
                    result = jac_value(self, value=list(atom_res.value.keys()))
                elif (op == "clear"):
                    result = jac_value(self, value=atom_res.value.clear())
                elif (op == "popitem"):
                    result = jac_value(self, value=list(
                        atom_res.value.popitem()))
                elif (op == "values"):
                    result = jac_value(self, value=list(
                        atom_res.value.values()))
                if (result):
                    if(len(kid) > 1):
                        self.rt_warn(
                            f"{op} does not take parameters, ignoring", kid[2])
                    return result
                if(len(kid) > 1):
                    args = self.run_expr_list(kid[2]).value
                    if (op == "pop"):
                        result = jac_value(
                            self, value=atom_res.value.pop(*args))
                    elif (op == "update"):
                        result = jac_value(
                            self, value=atom_res.value.update(*args))
                    if (result):
                        return result
            except Exception as e:
                self.rt_error(f'{e}', jac_ast)
            self.rt_error(f'Call to {op} is invalid.', jac_ast)

        return atom_res

    def run_list_built_in(self, jac_ast, atom_res):
        """
        list_built_in:
            KW_LENGTH
            | (TYP_LIST DBL_COLON | LIST_DBL_COLON) NAME (
                LPAREN expr_list RPAREN
            )?;
        """
        kid = self.set_cur_ast(jac_ast)
        if (kid[0].name == "KW_LENGTH"):
            if(isinstance(atom_res.value, list)):
                return jac_value(self, value=len(atom_res.value))
            else:
                self.rt_error(
                    f'Cannot get length of {atom_res.value}. Not List!',
                    kid[0])
                return jac_value(self, value=0)
        else:
            if(not self.rt_check_type(atom_res.value, [list], kid[0])):
                return atom_res
            kid = kid[1:]
            if(kid[0].name == "DBL_COLON"):
                kid = kid[1:]
            result = None
            op = kid[0].token_text()
            try:
                if (op == "reverse"):
                    result = jac_value(self, value=atom_res.value.reverse())
                elif (op == "copy"):
                    result = jac_value(self, value=atom_res.value.copy())
                elif (op == "sort"):
                    result = jac_value(self, value=atom_res.value.sort())
                elif (op == "clear"):
                    result = jac_value(self, value=atom_res.value.clear())
                elif (len(kid) < 2 and op == "pop"):
                    result = jac_value(self, value=atom_res.value.pop())
                if (result):
                    if(len(kid) > 1):
                        self.rt_warn(
                            f"{op} does not take parameters, ignoring", kid[2])
                    return result
                if(len(kid) > 1):
                    args = self.run_expr_list(kid[2]).value
                    if (op == "index"):
                        result = jac_value(
                            self, value=atom_res.value.index(*args))
                    elif (op == "append"):
                        result = jac_value(
                            self, value=atom_res.value.append(*args))
                    elif (op == "extend"):
                        result = jac_value(
                            self, value=atom_res.value.extend(*args))
                    elif (op == "insert"):
                        result = jac_value(
                            self, value=atom_res.value.insert(*args))
                    elif (op == "remove"):
                        result = jac_value(
                            self, value=atom_res.value.remove(*args))
                    elif (op == "count"):
                        result = jac_value(
                            self, value=atom_res.value.count(*args))
                    elif (op == "pop"):
                        result = jac_value(
                            self, value=atom_res.value.pop(*args))
                    if (result):
                        return result
            except Exception as e:
                self.rt_error(f'{e}', jac_ast)
            self.rt_error(f'Call to {op} is invalid.', jac_ast)
        return atom_res

    def run_string_built_in(self, jac_ast, atom_res):
        """
        string_built_in:
        (TYP_STRING DBL_COLON | STR_DBL_COLON) NAME (
                LPAREN expr_list RPAREN
        )?;
        """
        kid = self.set_cur_ast(jac_ast)
        if(not self.rt_check_type(atom_res.value, [str], kid[0])):
            return atom_res
        kid = kid[1:]
        if(kid[0].name == "DBL_COLON"):
            kid = kid[1:]
        result = None
        str_op = kid[0].token_text()
        try:
            if (str_op == "upper"):
                result = jac_value(self, value=atom_res.value.upper())
            elif (str_op == "lower"):
                result = jac_value(self, value=atom_res.value.lower())
            elif (str_op == "title"):
                result = jac_value(self, value=atom_res.value.title())
            elif (str_op == "capitalize"):
                result = jac_value(self, value=atom_res.value.capitalize())
            elif (str_op == "swap_case"):
                result = jac_value(self, value=atom_res.value.swapcase())
            elif (str_op == "is_alnum"):
                result = jac_value(self, value=atom_res.value.isalnum())
            elif (str_op == "is_alpha"):
                result = jac_value(self, value=atom_res.value.isalpha())
            elif (str_op == "is_digit"):
                result = jac_value(self, value=atom_res.value.isdigit())
            elif (str_op == "is_title"):
                result = jac_value(self, value=atom_res.value.istitle())
            elif (str_op == "is_upper"):
                result = jac_value(self, value=atom_res.value.isupper())
            elif (str_op == "is_lower"):
                result = jac_value(self, value=atom_res.value.islower())
            elif (str_op == "is_space"):
                result = jac_value(self, value=atom_res.value.isspace())
            elif (str_op == "load_json"):
                import json
                result = jac_value(self, value=json.loads(atom_res.value))
            elif (len(kid) < 2 and str_op == "split"):
                result = jac_value(self, value=atom_res.value.split())
            elif (len(kid) < 2 and str_op == "strip"):
                result = jac_value(self, value=atom_res.value.strip())
            elif (len(kid) < 2 and str_op == "lstrip"):
                result = jac_value(self, value=atom_res.value.lstrip())
            elif (len(kid) < 2 and str_op == "rstrip"):
                result = jac_value(self, value=atom_res.value.rstrip())
            if (result):
                if(len(kid) > 1):
                    self.rt_warn(
                        f"{str_op} does not take parameters, ignoring", kid[2])
                return result
            if(len(kid) > 1):
                args = self.run_expr_list(kid[2]).value
                if (str_op == "count"):
                    result = jac_value(self, value=atom_res.value.count(*args))
                elif (str_op == "find"):
                    result = jac_value(self, value=atom_res.value.find(*args))
                elif (str_op == "split"):
                    result = jac_value(self, value=atom_res.value.split(*args))
                elif (str_op == "join"):
                    result = jac_value(self, value=atom_res.value.join(*args))
                elif (str_op == "startswith"):
                    result = jac_value(
                        self, value=atom_res.value.startswith(*args))
                elif (str_op == "endswith"):
                    result = jac_value(
                        self, value=atom_res.value.endswith(*args))
                elif (str_op == "replace"):
                    result = jac_value(
                        self, value=atom_res.value.replace(*args))
                elif (str_op == "strip"):
                    result = jac_value(self, value=atom_res.value.strip(*args))
                elif (str_op == "lstrip"):
                    result = jac_value(
                        self, value=atom_res.value.lstrip(*args))
                elif (str_op == "rstrip"):
                    result = jac_value(
                        self, value=atom_res.value.rstrip(*args))
                if (result):
                    return result
        except Exception as e:
            self.rt_error(f'{e}', jac_ast)
        self.rt_error(f'Call to {str_op} is invalid.', jac_ast)
        return atom_res

    def run_node_edge_ref(self, jac_ast):
        """
        node_edge_ref:
            node_ref filter_ctx?
            | edge_ref (node_ref filter_ctx?)?;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'node_ref'):
            result = self.run_node_ref(kid[0])
            if(len(kid) > 1):
                result = self.run_filter_ctx(kid[1], result)
            return jac_value(self, value=result)

        elif (kid[0].name == 'edge_ref'):
            result = self.edge_to_node_jac_set(self.run_edge_ref(kid[0]))
            if(len(kid) > 1 and kid[1].name == 'node_ref'):
                nres = self.run_node_ref(kid[1])
                if(len(kid) > 2):
                    nres = self.run_filter_ctx(kid[2], nres)
                result = result * nres
            return jac_value(self, value=result)

    def run_node_ref(self, jac_ast, is_spawn=False):
        """
        node_ref: KW_NODE DBL_COLON NAME;
        """
        kid = self.set_cur_ast(jac_ast)
        if(not is_spawn):
            result = jac_set()
            if (len(kid) > 1):
                for i in self.viable_nodes().obj_list():
                    if (i.name == kid[2].token_text()):
                        result.add_obj(i)
            else:
                result += self.viable_nodes()
        else:
            result = self.parent().run_architype(
                kid[2].token_text(), kind='node', caller=self)
        return result

    def run_walker_ref(self, jac_ast):
        """
        walker_ref: KW_WALKER DBL_COLON NAME;
        """
        kid = self.set_cur_ast(jac_ast)
        return self.parent().spawn_walker(kid[2].token_text(), caller=self)

    def run_graph_ref(self, jac_ast):
        """
        graph_ref: KW_GRAPH DBL_COLON NAME;
        """
        kid = self.set_cur_ast(jac_ast)
        gph = self.parent().run_architype(
            kid[2].token_text(), kind='graph', caller=self)
        return gph

    def run_edge_ref(self, jac_ast, is_spawn=False):
        """
        edge_ref: edge_to | edge_from | edge_any;
        """
        kid = self.set_cur_ast(jac_ast)
        if(not is_spawn):
            return self.run_rule(kid[0])
        else:
            if(len(kid[0].kid) > 2):
                result = self.parent().run_architype(
                    kid[0].kid[2].token_text(), kind='edge',
                    caller=self)
                if(kid[0].kid[3].name == 'spawn_ctx'):
                    self.run_spawn_ctx(kid[0].kid[3], result)
                elif(kid[0].kid[3].name == 'filter_ctx'):
                    self.rt_error("Filtering not allowed here", kid[0].kid[3])
            else:
                result = edge(m_id=self._m_id, h=self._h,
                              kind='edge', name='generic')
            return result

    def run_edge_to(self, jac_ast):
        """
        edge_to:
            '-->'
            | '-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '->';
        """
        kid = self.set_cur_ast(jac_ast)
        result = jac_set()
        for i in self.current_node.outbound_edges() + \
                self.current_node.bidirected_edges():
            if (len(kid) > 2 and i.name != kid[2].token_text()):
                continue
            result.add_obj(i)
        if(len(kid) > 2 and kid[3].name == 'filter_ctx'):
            result = self.run_filter_ctx(kid[3], result)
        elif(len(kid) > 2 and kid[3].name == 'spawn_ctx'):
            self.rt_error("Assigning values not allowed here", kid[3])
        return result

    def run_edge_from(self, jac_ast):
        """
        edge_from:
            '<--'
            | '<-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '-';
        """
        kid = self.set_cur_ast(jac_ast)
        result = jac_set()
        for i in self.current_node.inbound_edges() + \
                self.current_node.bidirected_edges():
            if (len(kid) > 2 and i.name != kid[2].token_text()):
                continue
            result.add_obj(i)
        if(len(kid) > 2 and kid[3].name == 'filter_ctx'):
            result = self.run_filter_ctx(kid[3], result)
        elif(len(kid) > 2 and kid[3].name == 'spawn_ctx'):
            self.rt_error("Assigning values not allowed here", kid[3])
        return result

    def run_edge_any(self, jac_ast):
        """
        edge_any:
            '<-->'
            | '<-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '->';
        NOTE: these do not use strict bidirected semantic but any edge
        """
        kid = self.set_cur_ast(jac_ast)
        result = jac_set()
        for i in self.current_node.attached_edges():
            if (len(kid) > 2 and i.name != kid[2].token_text()):
                continue
            result.add_obj(i)
        if(len(kid) > 2 and kid[3].name == 'filter_ctx'):
            result = self.run_filter_ctx(kid[3], result)
        elif(len(kid) > 2 and kid[3].name == 'spawn_ctx'):
            self.rt_error("Assigning values not allowed here", kid[3])
        return result

    def run_list_val(self, jac_ast):
        """
        list_val: LSQUARE expr_list? RSQUARE;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[1].name == "expr_list"):
            return self.run_expr_list(kid[1])
        return jac_value(self, value=[])

    def run_index_slice(self, jac_ast, atom_res):
        """
        index_slice:
            LSQUARE expression RSQUARE
            | LSQUARE expression COLON expression RSQUARE;
        """
        kid = self.set_cur_ast(jac_ast)
        idx = self.run_expression(kid[1]).value
        if(type(idx) == str and idx not in atom_res.value.keys()):
            self.rt_error(f'Key {idx} not found in object/dict.', kid[1])
        if(kid[2].name == "RSQUARE"):
            if(not self.rt_check_type(idx, [int, str], kid[1])):
                self.rt_error(
                    f'Index of type {type(idx)} not valid. '
                    f'Indicies must be an integer or string!', kid[1])
                return atom_res
            atom_res.unwrap()
            try:
                return jac_value(self, ctx=atom_res.value, name=idx)
            except Exception:
                self.rt_error("List index out of range", kid[1])
                return atom_res
        else:
            end = self.run_expression(kid[3]).value
            if(not self.rt_check_type(idx, [int], kid[1]) or
               not self.rt_check_type(end, [int], kid[3])):
                self.rt_error('List slice range not valid. '
                              'Indicies must be an integers!', kid[1])
                return atom_res
            atom_res.unwrap()
            try:
                return jac_value(self, ctx=atom_res.value, name=idx, end=end)
            except Exception:
                self.rt_error("List slice out of range", kid[1])
                return atom_res

    def run_dict_val(self, jac_ast):
        """
        dict_val: LBRACE (kv_pair (COMMA kv_pair)*)? RBRACE;
        """
        kid = self.set_cur_ast(jac_ast)
        dict_res = {}
        for i in kid:
            if(i.name == 'kv_pair'):
                self.run_kv_pair(i, dict_res)
        return jac_value(self, value=dict_res)

    def run_kv_pair(self, jac_ast, obj):
        """
        kv_pair: STRING COLON expression;
        """
        kid = self.set_cur_ast(jac_ast)
        obj[parse_str_token(kid[0].token_text())
            ] = self.run_expression(kid[2]).value

    def run_spawn(self, jac_ast):
        """
        spawn: KW_SPAWN expression spawn_object;

        NOTE: spawn statements support locations that are either nodes or
        jac_sets
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[1].name == 'expression'):
            location = self.run_expression(kid[1]).value
            if(isinstance(location, node)):
                return self.run_spawn_object(kid[2], location)
            elif(isinstance(location, jac_set)):
                res = []
                for i in location.obj_list():
                    res.append(self.run_spawn_object(kid[2], i))
                return jac_value(self, value=res)
            else:
                self.rt_error(
                    f'Spawn can not occur on {type(location)}!', kid[1])
        else:
            return self.run_spawn_object(kid[1], None)

    def run_spawn_object(self, jac_ast, location):
        """
        spawn_object: node_spawn | walker_spawn;
        """
        kid = self.set_cur_ast(jac_ast)
        return self.run_rule(kid[0], location)

    def run_node_spawn(self, jac_ast, location):
        """
        node_spawn: edge_ref? node_ref spawn_ctx?;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == 'node_ref'):
            ret_node = self.run_node_ref(kid[0], is_spawn=True)
        else:
            use_edge = self.run_edge_ref(kid[0], is_spawn=True)
            ret_node = self.run_node_ref(kid[1], is_spawn=True)
            direction = kid[0].kid[0].name
            if (direction == 'edge_from'):
                location.attach_inbound(ret_node, [use_edge])
            elif (direction == 'edge_to'):
                location.attach_outbound(ret_node, [use_edge])
            else:
                location.attach_bidirected(ret_node, [use_edge])
        if (kid[-1].name == 'spawn_ctx'):
            self.run_spawn_ctx(kid[-1], ret_node)
        return jac_value(self, value=ret_node)

    def run_walker_spawn(self, jac_ast, location):
        """
        walker_spawn: walker_ref spawn_ctx?;
        """
        kid = self.set_cur_ast(jac_ast)
        walk = self.run_walker_ref(kid[0])
        walk.prime(location)
        if(len(kid) > 1):
            self.run_spawn_ctx(kid[1], walk)
        walk.run()
        ret = jac_value(self, value=walk.anchor_value())
        ret.unwrap()
        self.report += walk.report
        self.runtime_errors += walk.runtime_errors
        walk.destroy()
        return ret

    def run_graph_spawn(self, jac_ast, location):
        """
        graph_spawn: edge_ref graph_ref;
        """
        kid = self.set_cur_ast(jac_ast)
        use_edge = self.run_edge_ref(kid[0], is_spawn=True)
        result = self.run_graph_ref(kid[1])
        direction = kid[0].kid[0].name
        if (direction == 'edge_from'):
            location.attach_inbound(result, [use_edge])
        elif (direction == 'edge_to'):
            location.attach_outbound(result, [use_edge])
        else:
            location.attach_bidirected(result, [use_edge])
        return jac_value(self, value=result)

    def run_spawn_ctx(self, jac_ast, obj):
        """
        spawn_ctx: LPAREN (spawn_assign (COMMA spawn_assign)*)? RPAREN;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if (i.name == 'spawn_assign'):
                self.run_spawn_assign(i, obj)

    def run_filter_ctx(self, jac_ast, obj):
        """
        filter_ctx:
                LPAREN (filter_compare (COMMA filter_compare)*)? RPAREN;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = jac_set()
        for i in obj.obj_list():
            for j in kid:
                if (j.name == 'filter_compare'):
                    if(self.run_filter_compare(j, i)):
                        ret.add_obj(i)
        return ret

    def run_spawn_assign(self, jac_ast, obj):
        """
        spawn_assign: NAME EQ expression;
        """
        kid = self.set_cur_ast(jac_ast)
        name = kid[0].token_text()
        result = self.run_expression(kid[-1]).value
        dest = jac_value(self, ctx=obj, name=name, value=result)
        if(obj.j_type == 'walker'):
            dest.write(kid[0], force=True)
        else:
            dest.write(kid[0])

    def run_filter_compare(self, jac_ast, obj):
        """
        filter_compare: NAME cmp_op expression;
        """
        kid = self.set_cur_ast(jac_ast)
        name = kid[0].token_text()
        if(name in obj.context.keys()):
            result = self.run_expression(kid[-1])
            return self.run_cmp_op(
                kid[1], jac_value(self, ctx=obj, name=name),
                result).value
        else:
            self.rt_error(f'{name} not present in object', kid[0])
            return False

    def run_any_type(self, jac_ast):
        """
        any_type:
            TYP_STRING
            | TYP_INT
            | TYP_FLOAT
            | TYP_LIST
            | TYP_DICT
            | TYP_BOOL
            | KW_NODE
            | KW_EDGE
            | KW_TYPE;
        """
        kid = self.set_cur_ast(jac_ast)
        if(kid[0].name == "TYP_STRING"):
            return jac_value(self, value=str)
        elif(kid[0].name == "TYP_INT"):
            return jac_value(self, value=int)
        elif(kid[0].name == "TYP_FLOAT"):
            return jac_value(self, value=float)
        elif(kid[0].name == "TYP_LIST"):
            return jac_value(self, value=list)
        elif(kid[0].name == "TYP_DICT"):
            return jac_value(self, value=dict)
        elif(kid[0].name == "TYP_BOOL"):
            return jac_value(self, value=bool)
        elif(kid[0].name == "KW_NODE"):
            return jac_value(self, value=node)
        elif(kid[0].name == "KW_EDGE"):
            return jac_value(self, value=edge)
        elif(kid[0].name == "KW_TYPE"):
            return jac_value(self, value=type)
        else:
            self.rt_error('Unrecognized type', kid[0])

    # Helper Functions ##################

    def call_ability(self, nd, name, act_list):
        m = interp(parent_override=self.parent(), caller=self)
        m.push_scope(jac_scope(parent=nd,
                               has_obj=nd,
                               action_sets=[nd.activity_action_ids]))
        m._jac_scope.inherit_agent_refs(self._jac_scope)
        m.run_code_block(jac_ir_to_ast(
            act_list.get_obj_by_name(name).value))
        self.report += m.report
        self.runtime_errors += m.runtime_errors

    def run_rule(self, jac_ast, *args):
        """Helper to run rule if exists in execution context"""
        # try:
        # if(not hasattr(self, f'run_{jac_ast.name}')):
        #     self.rt_error(
        #         f'This scope cannot execute the statement '
        #         f'"{jac_ast.get_text()}" of type {jac_ast.name}',
        #         jac_ast)
        #     return

        return getattr(self, f'run_{jac_ast.name}')(jac_ast, *args)
        # except Exception as e:
        #     self.rt_error(
        #         f"Cannot execute this type of code here! {e}", jac_ast)
        #     return None



"""
Interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.utils.utils import logger
from jaseci.actions.live_actions import live_actions, load_preconfig_actions
# from jaseci.actions.find_action import find_action
from jaseci.element.element import element

from jaseci.jac.jac_set import jac_set
from jaseci.jac.machine.jac_scope import jac_scope


class machine_state():
    """Shared interpreter class across both sentinels and walkers"""

    def __init__(self, parent_override=None, caller=None):
        self.report = []
        self.runtime_errors = []
        self._parent_override = parent_override
        if(not isinstance(self, element) and caller):
            self._m_id = caller._m_id
            self._h = caller._h
        self._scope_stack = [None]
        self._jac_scope = None
        self._loop_ctrl = None
        self._stopped = None
        self._assign_mode = False
        self._loop_limit = 10000
        self._cur_jac_ast = None

    def parent(self):
        if(self._parent_override):
            return self._parent_override
        else:
            return element.parent(self)

    def reset(self):
        self.report = []
        self.runtime_errors = []
        self._scope_stack = [None]
        self._jac_scope = None
        self._loop_ctrl = None
        self._stopped = None

    def push_scope(self, scope: jac_scope):
        self._scope_stack.append(scope)
        self._jac_scope = scope

    def pop_scope(self):
        self._scope_stack.pop()
        self._jac_scope = self._scope_stack[-1]

    def set_cur_ast(self, jac_ast):
        self._cur_jac_ast = jac_ast
        return jac_ast.kid

    # Helper Functions ##################

    def obj_set_to_jac_set(self, obj_set):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = jac_set()
        for i in obj_set:
            ret.add_obj(i)
        return ret

    def edge_to_node_jac_set(self, edge_set):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = jac_set()
        for i in edge_set.obj_list():
            ret.add_obj(i.opposing_node(self.current_node))
        return ret

    def get_builtin_action(self, func_name, jac_ast=None):
        """
        Takes reference to action attr, finds the built in function
        and returns new name used as hook by action class
        """
        if(func_name not in live_actions.keys()):
            load_preconfig_actions(self._h)
        if(func_name not in live_actions.keys()):
            self.rt_error(f"Builtin action not found - {func_name}", jac_ast)
        return func_name

    def jac_exception(self, e: Exception, jac_ast):
        return {'type': type(e).__name__,
                'mod': jac_ast.mod_name,
                'msg': str(e),
                'args': e.args,
                'line': jac_ast.line,
                'col': jac_ast.column}

    def rt_log_str(self, msg, jac_ast=None):
        """Generates string for screen output"""
        name = self.name if hasattr(self, 'name') else 'blank'
        if(jac_ast):
            msg = f'{jac_ast.mod_name}:{name} - line {jac_ast.line}, ' + \
                f'col {jac_ast.column} - rule {jac_ast.name} - {msg}'
        else:
            msg = f'{msg}'
        return msg

    def rt_warn(self, error, jac_ast=None):
        """Prints runtime error to screen"""
        error = self.rt_log_str(error, jac_ast)
        logger.warning(
            str(error)
        )
        self.runtime_errors.append(error)

    def rt_error(self, error, jac_ast=None):
        """Prints runtime error to screen"""
        error = self.rt_log_str(error, jac_ast)
        logger.error(
            str(error)
        )
        self.runtime_errors.append(error)

    def rt_info(self, msg, jac_ast=None):
        """Prints runtime info to screen"""
        logger.info(
            str(self.rt_log_str(msg, jac_ast))
        )

    def rt_check_type(self, obj, typ, jac_ast=None):
        """Prints error if type mismatach"""
        if(not isinstance(typ, list)):
            typ = [typ]
        for i in typ:
            if (isinstance(obj, i)):
                return True
        self.rt_error(f'Incompatible type for object '
                      f'{obj} - {type(obj).__name__}, '
                      f'expecting {typ}', jac_ast)
        return False



from jaseci.api.interface import interface
import stripe

stripe_test_key = "sk_test_4eC39HqLyjWDarjtT1zdp7dc"


class stripe_api():
    """
    Stripe APIs
    Set of APIs to expose jaseci stripe management
    """

    def __init__(self):
        stripe.api_key = self._h.resolve_glob('STRIPE_KEY', stripe_test_key)

    @interface.admin_api()
    def stripe_product_create(self, name: str = "VIP Plan",
                              description: str = "Plan description"):
        """ create product """
        try:
            return stripe.Product.create(
                namex=name,
                descriptionx=description
            )
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_product_price_set(self, productId: str,
                                 amount: float = 50,
                                 interval: str = "month"):
        """ modify product price """
        try:
            return stripe.Price.create(
                unit_amount=amount,
                currency="usd",
                recurring={"interval": interval},
                product=productId,
            )
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_product_list(self, detalied: bool = True):
        """ retrieve all producs """
        if(not detalied):
            try:
                return stripe.Product.list(active=True)
            except Exception as e:
                return {"message": str(e)}
        else:  # retrieve product price
            try:
                return stripe.Product.list()
            except Exception as e:
                return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_create(self, paymentId: str,
                               name: str = "cristopher evangelista",
                               email: str = "imurbatman12@gmail.com",
                               description: str = "Myca subscriber"):
        """ create customer """
        try:
            return stripe.Customer.create(
                name=name,
                email=email,
                description=description,
                payment_method=paymentId,
                invoice_settings={"default_payment_method": paymentId}
            )
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_get(self, customerId: str):
        """ retrieve customer information """
        try:
            return stripe.Customer.retrieve(customerId)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_add(self, paymentMethodId: str,
                                    customerId: str):
        """ add customer payment method """
        try:
            paymentMethods = \
                self.stripe_retrieve_customer_payment_methods(
                    customerId)

            paymentMethod = stripe.PaymentMethod.attach(
                paymentMethodId, customer=customerId)

            if(len(paymentMethods.data) == 0):
                self.stripe_update_default_payment_method(
                    customerId, paymentMethodId)

            paymentMethod.is_default = len(paymentMethods.data) == 0

            return paymentMethod

        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_delete(self, paymentMethodId: str):
        """ remove customer payment method """
        try:
            return stripe.PaymentMethod.detach(paymentMethodId)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_get(self, customerId: str):
        """ get customer list of payment methods """
        try:
            return stripe.PaymentMethod.list(
                customer=customerId,
                type="card",
            )
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_customer_payment_default(self, customerId: str,
                                        paymentMethodId: str):
        """ update default payment method of customer """
        try:
            setting = {"default_payment_method": paymentMethodId}
            return stripe.Customer.modify(customerId,
                                          invoice_settings=setting)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_subscription_create(self, paymentId: str,
                                   name: str, email: str,
                                   priceId: str, customerId: str):
        """
        create customer subscription
        TODO: name and email parameters not used!
        """
        try:
            # attach payment method to customer
            self.stripe_add_customer_payment_methods(
                paymentId, customerId)

            # set card to default payment method
            self.stripe_update_default_payment_method(
                customerId, paymentId)

            subscription = stripe.Subscription.create(
                customer=customerId,
                items=[
                    {"price": priceId},
                ],
                trial_period_days=30
            )

            subscription.payment_method = paymentId

            return subscription
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_subscription_delete(self, subscriptionId: str):
        """ cancel customer subscription """
        try:
            return stripe.Subscription.delete(subscriptionId)
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_subscription_get(self, customerId: str):
        """ retrieve customer subcription """
        try:
            subscription = stripe.Subscription.list(customer=customerId)

            if (len(subscription.data) == 0):
                return {"status": "inactive", "message":
                        "Customer has no subscription"}

            subscription.pop("items", None)

            return subscription.data[0]
        except Exception as e:
            return {"message": str(e)}

    @interface.admin_api()
    def stripe_invoices_list(self, customerId: str,
                             subscriptionId: str,
                             limit: int = 10,
                             lastItem: str = ""):
        """ retrieve customer list of invoices """
        try:
            if(lastItem != ''):
                invoices = stripe.Invoice.list(
                    customer=customerId, limit=limit, starting_after=lastItem,
                    subscription=subscriptionId)
            else:
                invoices = stripe.Invoice.list(
                    customer=customerId, limit=limit,
                    subscription=subscriptionId)
            return invoices
        except Exception as e:
            return {"message": str(e)}



"""
Admin Global api functions as a mixin
"""
from jaseci.api.interface import interface
import jaseci.actions.live_actions as lact
import json


class actions_api():
    """
    Admin global APIs
    """

    @interface.admin_api(cli_args=['file'])
    def actions_load_local(self, file: str):
        """
        Hot load a python module and assimlate any Jaseci Actions
        """
        success = lact.load_local_actions(file)
        if(success):
            cur_config = self.config_get('ACTION_SETS')
            if(cur_config and (not isinstance(cur_config, list))):
                config = json.loads(cur_config)
                if file not in config['local']:
                    config['local'].append(file)
                    self.config_set('ACTION_SETS', json.dumps(config))
            else:
                self.config_set('ACTION_SETS',
                                json.dumps({'local': [file], 'remote': []}))
        return {"success": success}

    @interface.admin_api(cli_args=['url'])
    def actions_load_remote(self, url: str):
        """
        Hot load an actions set from live pod at URL
        """
        success = lact.load_remote_actions(url)
        if(success):
            cur_config = self.config_get('ACTION_SETS')
            if(cur_config and (not isinstance(cur_config, list))):
                config = json.loads(cur_config)
                if url not in config['remote']:
                    config['remote'].append(url)
                    self.config_set('ACTION_SETS', json.dumps(config))
            else:
                self.config_set('ACTION_SETS',
                                json.dumps({'local': [], 'remote': [url]}))
        return {"success": success}

    # @interface.admin_api()
    # def actions_get(self, name: str, value: str):
    #     """
    #     """

    @interface.admin_api()
    def actions_list(self, name: str = ''):
        """
        List all live jaseci actions
        """
        actions = list(lact.live_actions.keys())
        if(len(name)):
            actions = list(filter(lambda a: a.startswith(name), actions))
        return actions

    # @interface.admin_api()
    # def actions_delete(self, name: str, value: str):
    #     """
    #     """



"""
Object api as a mixin
"""
from jaseci.api.interface import interface
from jaseci.element.element import element


class object_api():
    """Object APIs for generalized operations on Jaseci objects

    ...
    """

    @interface.private_api(cli_args=['name'])
    def global_get(self, name: str):
        """
        Get a global var
        """
        return {'value': self._h.get_glob(name)}

    @interface.private_api(cli_args=['obj'])
    def object_get(self, obj: element, depth: int = 0,
                   detailed: bool = False):
        """Returns object details for any Jaseci object.

        """
        return obj.serialize(deep=depth, detailed=detailed)

    @interface.private_api(cli_args=['obj'])
    def object_perms_get(self, obj: element):
        """Returns object access mode for any Jaseci object.

        """
        return {'access': obj.j_access}

    @interface.private_api(cli_args=['obj'])
    def object_perms_set(self, obj: element, mode: str):
        """Sets object access mode for any Jaseci object.

        """
        valid_perms = ['public', 'private', 'read_only']
        ret = {}
        if(mode not in valid_perms):
            ret['success'] = False
            ret['response'] = f'{mode} not valid, must be in {valid_perms}'
        else:
            getattr(obj, 'make_'+mode)()
            ret['success'] = True
            ret['response'] = f'{obj} set to {mode}'
        return ret

    @interface.private_api(cli_args=['obj'])
    def object_perms_grant(self, obj: element, mast: element,
                           read_only: bool = False):
        """Grants another user permissions to access a Jaseci object.

        """
        granted = obj.give_access(mast, read_only=read_only)
        ret = {'success': granted}
        if(granted):
            ret['response'] = f'Access to {obj} granted to {mast}'
        else:
            ret['response'] = f'Cannot grant {mast} access to {obj}'
        return ret

    @interface.private_api(cli_args=['obj'])
    def object_perms_revoke(self, obj: element, mast: element):
        """Remove permissions for user to access a Jaseci object.

        """
        revoked = obj.remove_access(mast)
        ret = {'success': revoked}
        if(revoked):
            ret['response'] = f'Access to {obj} revoked from {mast}'
        else:
            ret['response'] = f'{mast} did not have access to {obj}'
        return ret



"""
Architype api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.actor.architype import architype
from jaseci.actor.sentinel import sentinel
from jaseci.utils.utils import b64decode_str


class architype_api():
    """Architype APIs for creating and managing Jaseci architypes

    The architype set of APIs allow for the addition and removing of
    architypes. Given a Jac implementation of an architype these APIs are
    designed for creating, compiling, and managing architypes that can be
    used by Jaseci. There are two ways to add an architype to Jaseci, either
    through the management of sentinels using the sentinel API, or by
    registering independent architypes with these architype APIs. These
    APIs are also used for inspecting and managing existing arichtypes that
    a Jaseci instance is aware of.
    """

    @interface.private_api(cli_args=['code'])
    def architype_register(self, code: str, encoded: bool = False,
                           snt: sentinel = None):
        """Create an architype based on the code passed and return object.

        This register API allows for the creation or replacement/update of
        an architype that can then be used by walkers in their interactions
        of graphs. The code argument takes Jac source code for the single
        architype. To load multiple architypes and walkers at the same time,
        use sentinel register API.

        Args:
            code (str): The text (or filename) for an architypes Jac code
            encoded (bool): True/False flag as to whether code is encode
                in base64
            snt (uuid): The UUID of the sentinel to be the owner of this
                architype

        Returns:
            json: Fields include
                'architype': Architype object if created otherwise null
                'success': True/False whether register was successful
                'errors': List of errors if register failed
                'response': Message on outcome of register call
        """
        ret = {'architype': None, 'success': False, 'errors': []}
        if (encoded):
            code = b64decode_str(code)
        arch = snt.register_architype(code)
        if(arch):
            self.extract_arch_aliases(snt, arch)
            ret['architype'] = arch.serialize()
            ret['success'] = True
            ret['response'] = f'Successfully created {arch.name} architype'
        else:
            ret['errors'] = snt.errors
            ret['response'] = f'Errors occured'
        return ret

    @interface.private_api()
    def architype_get(self, arch: architype, mode: str = 'default',
                      detailed: bool = False):
        """Get an architype rendered with specific mode

        Args:
            arch (uuid): The architype being accessed
            mode (str): Valid modes: {default, code, ir, }
            detailed (bool): Flag to give summary or complete set of fields

        Returns:
            json: Fields include (depends on mode)
                'code': Formal source code for architype
                'ir': Intermediate representation of architype
                'architype': Architype object print
        """
        if(mode == 'code'):
            return {'code': arch._jac_ast.get_text()}
        elif(mode == 'ir'):
            return {'ir': arch.ir_dict()}
        else:
            return {'architype': arch.serialize(detailed=detailed)}

    @interface.private_api(cli_args=['code'])
    def architype_set(self, arch: architype, code: str,
                      mode: str = 'default'):
        """Set code/ir for a architype

        Args:
            arch (uuid): The architype being set
            code (str): The text (or filename) for an architypes Jac code/ir
            mode (str): Valid modes: {default, code, ir, }

        Returns:
            json: Fields include (depends on mode)
                'success': True/False whether set was successful
                'errors': List of errors if set failed
                'response': Message on outcome of set call
        """
        if(mode == 'code' or mode == 'default'):
            arch.register(code)
        elif(mode == 'ir'):
            arch.apply_ir(code)
        else:
            return {'response': f'Invalid mode to set {arch}',
                    'success': False, 'errors': []}
        if(arch.is_active):
            return {'response': f'{arch} registered and active!',
                    'success': True, 'errors': []}
        else:
            return {'response': f'{arch} registered and active!',
                    'success': True,
                    'errors': arch.errors}

    @interface.private_api()
    def architype_list(self, snt: sentinel = None, detailed: bool = False):
        """List architypes known to sentinel

        Args:
            snt (uuid): The sentinel for which to list its architypes
            detailed (bool): Flag to give summary or complete set of fields

        Returns:
            json: List of architype objects
        """
        archs = []
        for i in snt.arch_ids.obj_list():
            archs.append(i.serialize(detailed=detailed))
        return archs

    @interface.private_api(cli_args=['arch'])
    def architype_delete(self, arch: architype, snt: sentinel = None):
        """Permanently delete sentinel with given id

        Args:
            arch (uuid): The architype being set
            snt (uuid): The sentinel for which to list its architypes

        Returns:
            json: Fields include (depends on mode)
                'success': True/False whether command was successful
                'response': Message on outcome of command
        """

        if(arch.jid not in snt.arch_ids):
            return {'response': f'Architype {arch} not in sentinel {snt}',
                    'success': False}
        self.remove_arch_aliases(snt, arch)
        archid = arch.jid
        snt.arch_ids.destroy_obj(arch)
        return {'response': f'Architype {archid} successfully deleted',
                'success': True}



from unittest import TestCase
from jaseci.utils.utils import TestCaseHelper
from jaseci.jsctl import jsctl
from click.testing import CliRunner
import json


class jsctl_test(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def call(self, cmd):
        res = CliRunner(mix_stderr=False).invoke(jsctl.jsctl,
                                                 ["-m"]+cmd.split(' '))
        # self.log(res.stdout)
        # self.log(res.stderr)
        # self.log(res.exception)
        return res.stdout

    def call_cast(self, cmd):
        ret = self.call(cmd)
        # self.log(ret)
        return json.loads(ret)

    def call_split(self, cmd):
        ret = self.call(cmd)
        # self.log(ret)
        return ret.split('\n')

    def tearDown(self):
        jsctl.session["master"]._h.clear_mem_cache()
        super().tearDown()

    def test_jsctl_extract_tree(self):
        out = jsctl.extract_api_tree()
        self.assertIn("graph", out)
        self.assertIn("walker", out)
        self.assertIn("sentinel", out)
        self.assertIn("alias", out)
        self.assertIn("config", out)
        self.assertNotIn("jadc", out)

    def test_help_screen(self):
        r = self.call('--help')
        self.assertIn('Specify filename', r)
        self.assertIn("Group of `graph` commands", r)
        r = self.call('sentinel --help')
        self.assertIn("Group of `sentinel active` commands", r)

    def test_jsctl_create_graph_mem_only(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/ll.jac -name ll -set_active true")
        self.assertEqual(len(self.call_cast('graph list')), 1)
        self.call('graph create')
        self.call('graph create')
        self.call('graph create')
        self.assertEqual(len(self.call_cast('graph list')), 4)

    def test_jsctl_carry_forward(self):
        self.call(
            "actions load local ../jskit/infer.py")
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/ll.jac -name ll -set_active true")
        self.call("graph create -set_active true")
        self.call("walker run init")
        self.call("walker run gen_rand_life")
        r = self.call_cast("walker run get_gen_day")
        self.assertGreater(len(r['report']), 3)

    def test_jsctl_dot(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/ll.jac -name ll -set_active true")
        self.call("graph create -set_active true")
        self.call("walker run init")
        self.call("walker run gen_rand_life")
        r = self.call("graph get -mode dot")
        self.assertIn("test test test", r)
        self.assertIn('"n0" -> "n', r)
        self.assertIn('week="', r)

    def test_jsctl_aliases(self):
        """Tests that alias mapping api works"""
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/zsb.jac -name zsb -set_active true")
        gph_id = self.call_cast('graph create')['jid']
        snt_id = self.call_cast('sentinel list')[0]['jid']
        self.call(f'alias register s -value {snt_id}')
        self.call(f'alias register g -value {gph_id}')
        self.assertEqual(len(self.call_cast('graph get -gph g')), 1)
        self.call(f'walker run init -snt s -nd g')
        self.assertEqual(len(self.call_cast('graph get -gph g')), 3)
        self.call(f'alias clear')
        self.assertEqual(len(self.call_cast(f'alias list').keys()), 0)

    def test_jsctl_auto_aliases(self):
        """Tests that auto alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb "
            "-set_active true")
        aliases = self.call_cast('alias list')
        self.assertGreater(len(aliases), 20)
        self.assertIn('zsb:architype:bot', aliases.keys())
        self.assertIn('zsb:walker:similar_questions', aliases.keys())
        self.assertIn('sentinel:zsb', aliases.keys())
        self.assertIn('zsb:walker:create_nugget', aliases.keys())

    def test_jsctl_auto_aliases_delete(self):
        """Tests that auto removing alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb "
            "-set_active true")
        num = len(self.call_cast('alias list'))

        self.call('architype delete zsb:architype:bot')
        self.call('walker delete zsb:walker:similar_questions')
        self.call('walker delete zsb:walker:create_nugget')
        self.assertEqual(num-3, len(self.call_cast('alias list')))

    def test_jsctl_config_cmds(self):
        """Tests that config commands works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb "
            "-set_active true")
        self.call(
            f'config set CONFIG_EXAMPLE -value TEST -do_check False')
        self.call(f'config set APPLE -value Grape2 -do_check False')
        self.call(f'config set "Banana" -value "Grape" -do_check False')
        self.call(f'config set "Pear" -value "Banana" -do_check False')
        r = self.call('config get APPLE -do_check False')
        self.assertEqual(r.strip(), 'Grape2')
        r = self.call_cast('config list')
        self.assertEqual(len(r), 1)

    def test_jsctl_default_snt_setting(self):
        """Tests that alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/teststest.jac -name zsb "
            "-set_active true")
        snt_id = self.call_cast('sentinel list')[0]['jid']
        self.call(f'sentinel active set {snt_id}')
        self.call(f'sentinel active get')
        self.assertEqual(len(self.call_cast(f'walker list')), 2)

    def test_jsctl_init_auto_called(self):
        """Tests that alias mapping api works"""
        num = len(self.call_cast('graph get'))
        self.assertEqual(num, 3)

    def test_jsctl_master_defaults(self):
        """Tests that alias mapping api works"""
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb "
            "-set_active true")
        gph_id = self.call_cast('graph create')['jid']
        snt_id = self.call_cast('sentinel list')[0]['jid']
        self.call(f'sentinel active set -snt {snt_id}')
        self.call(f'graph active set -gph {gph_id}')
        self.assertEqual(len(self.call_cast('graph get')), 1)
        self.call(f'walker run init')
        self.assertEqual(len(self.call_cast('graph get')), 3)

    def test_public_apis_present(self):
        r = self.call('walker --help')
        self.assertIn('summon', r)
        self.assertIn('namespace key', r)

    def test_public_apis_walker_summon_auth(self):
        self.call(
            "sentinel register jaseci/jsctl/tests/zsb.jac -name zsb "
            "-set_active true")
        r = self.call_cast('walker get zsb:walker:pubinit -mode keys')
        key = r['anyone']
        r = self.call_cast('alias list')
        walk = r['zsb:walker:pubinit']
        nd = r['active:graph']
        r = self.call_cast(f'walker summon {walk} -key {key} -nd {nd}')
        self.assertEqual(len(r['report']), 0)
        key = 'aaaaaaaa'
        r = self.call_cast(f'walker summon {walk} -key {key} -nd {nd}')
        self.assertFalse(r['success'])

    def test_jsctl_import(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/ll_base.jac -set_active true")
        self.logger_off()
        self.call("walker run init")
        self.call("walker run gen_rand_life")
        r = self.call_cast("walker run get_gen_day")
        self.assertGreater(len(r['report']), 3)

    def test_jsctl_import_filters(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/base.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r['report']), 8)

    def test_jsctl_import_filters1(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/base1.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r['report']), 8)

    def test_jsctl_import_filters2(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/base2.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r['report']), 8)

    def test_jsctl_import_fails_when_incomplete(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/base3.jac -set_active true")
        r = self.call_cast("walker run init")
        self.assertTrue('success' in r.keys())
        self.assertFalse(r['success'])

    def test_jsctl_run_tests(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/teststest.jac")
        r = self.call_split("sentinel test")
        self.assertTrue(r[0].startswith('Testing "assert should be'))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_run_tests_detailed(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/teststest.jac")
        r = self.call_split("sentinel test -detailed true")
        self.assertEqual(len(r), 27)

    def test_jsctl_jac_build(self):
        import os
        if(os.path.exists('jaseci/jsctl/tests/teststest.jir')):
            os.remove('jaseci/jsctl/tests/teststest.jir')
            self.assertFalse(os.path.exists(
                'jaseci/jsctl/tests/teststest.jir'))
        self.call(
            "jac build jaseci/jsctl/tests/teststest.jac")
        self.assertGreater(os.path.getsize(
            'jaseci/jsctl/tests/teststest.jir'), 50000)

    def test_jsctl_jac_test(self):
        r = self.call_split(
            "jac test jaseci/jsctl/tests/teststest.jac")
        self.assertTrue(r[0].startswith('Testing "assert should be'))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_jac_test_jir(self):
        r = self.call_split(
            "jac test jaseci/jsctl/tests/teststest.jir")
        self.assertTrue(r[0].startswith('Testing "assert should be'))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_jac_run(self):
        r = self.call_cast(
            "jac run jaseci/jsctl/tests/teststest.jac")
        self.assertEqual(r['report'], [{}, 4])

    def test_jsctl_jac_run_jir(self):
        r = self.call_cast(
            "jac run jaseci/jsctl/tests/teststest.jir")
        self.assertEqual(r['report'], [{}, 4])

    def test_jsctl_jac_run_jir_walk(self):
        r = self.call_cast(
            "jac run jaseci/jsctl/tests/teststest.jir -walk alt_init")
        self.assertEqual(r['report'], [7])

    def test_jsctl_sentinel_set(self):
        self.call(
            "sentinel register "
            "jaseci/jsctl/tests/teststest.jac -set_active true")
        self.call(
            "sentinel set "
            "jaseci/jsctl/tests/base.jac")
        r = self.call_cast("walker run init")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r['report']), 8)

    def test_jsctl_action_list(self):
        r = self.call_cast(
            "actions list")
        self.assertIn("std.out", r)

    def test_jsctl_master_self(self):
        r = self.call_cast(
            "master self")
        self.assertIn("j_type", r.keys())
        a = r['jid']
        r = self.call_cast(
            "master active get")
        self.assertIn("j_type", r.keys())
        b = r['jid']
        self.assertEqual(a, b)



"""
Walker api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.actor.walker import walker
from jaseci.graph.node import node
from jaseci.actor.sentinel import sentinel
from jaseci.utils.utils import b64decode_str
from jaseci.utils.id_list import id_list


class walker_api():
    """
    Walker APIs
    """

    def __init__(self):
        self.spawned_walker_ids = id_list(self)

    @interface.public_api(cli_args=['wlk'])
    def walker_summon(self, key: str, wlk: walker, nd: node,
                      ctx: dict = {}, global_sync: bool = True):
        """
        Public api for running walkers, namespace key must be provided
        along with the walker id and node id
        """
        if(key not in wlk.namespace_keys().values()):
            return self.bad_walk_response(
                ['Not authorized to execute this walker'])
        if(global_sync):  # Test needed
            self.sync_walker_from_global_sent(wlk)
        walk = wlk.duplicate()
        walk.refresh()
        walk.prime(nd, prime_ctx=ctx)
        res = walk.run()
        walk.destroy()
        return res

    @interface.private_api(cli_args=['code'])
    def walker_register(self, snt: sentinel = None,
                        code: str = '', encoded: bool = False):
        """
        Create blank or code loaded walker and return object
        """
        if (encoded):
            code = b64decode_str(code)
        walk = snt.register_walker(code)
        if(walk):
            self.extract_wlk_aliases(snt, walk)
            return walk.serialize()
        else:
            return [f'Walker not created, invalid code!']

    @interface.private_api(cli_args=['wlk'])
    def walker_get(self, wlk: walker, mode: str = 'default',
                   detailed: bool = False):
        """
        Get a walker rendered with specific mode
        Valid modes: {default, code, ir, keys, }
        """
        if(mode == 'code'):
            return wlk._jac_ast.get_text()
        elif(mode == 'ir'):
            return wlk.ir_dict()
        elif(mode == 'keys'):
            return wlk.namespace_keys()
        else:
            return wlk.serialize(detailed=detailed)

    @interface.private_api(cli_args=['wlk'])
    def walker_set(self, wlk: walker, code: str,
                   mode: str = 'default'):
        """
        Set code/ir for a walker
        Valid modes: {code, ir, }
        """
        if(mode == 'code' or mode == 'default'):
            wlk.register(code)
        elif(mode == 'ir'):
            wlk.apply_ir(code)
        else:
            return [f'Invalid mode to set {wlk}']
        if(wlk.is_active):
            return [f'{wlk} registered and active!']
        else:
            return [f'{wlk} code issues encountered!']

    @interface.private_api()
    def walker_list(self, snt: sentinel = None, detailed: bool = False):
        """
        List walkers known to sentinel
        """
        walks = []
        for i in snt.walker_ids.obj_list():
            walks.append(i.serialize(detailed=detailed))
        return walks

    @interface.private_api(cli_args=['wlk'])
    def walker_delete(self, wlk: walker, snt: sentinel = None):
        """
        Permanently delete walker with given id
        """
        self.remove_wlk_aliases(snt, wlk)
        wlkid = wlk.id
        snt.walker_ids.destroy_obj(wlk)
        return [f'Walker {wlkid} successfully deleted']

    @interface.private_api(cli_args=['name'])
    def walker_spawn_create(self, name: str, snt: sentinel = None):
        """
        Creates new instance of walker and returns new walker object
        """
        wlk = snt.spawn_walker(name, caller=self)
        if(wlk):
            if(self.spawned_walker_ids.has_obj_by_name(name)):
                self.spawned_walker_ids.destroy_obj_by_name(name)
            self.spawned_walker_ids.add_obj(wlk)
            self.alias_register(f'spawned:walker:{name}', wlk.jid)
            return wlk.serialize()
        else:
            return [f'Walker not found!']

    @interface.private_api(cli_args=['name'])
    def walker_spawn_delete(self, name: str):
        """
        Delete instance of walker
        """
        if(self.spawned_walker_ids.has_obj_by_name(name)):
            self.spawned_walker_ids.destroy_obj_by_name(name)
            self.alias_delete(f'spawned:walker:{name}')
            return [f'Walker {name} deteled!']
        else:
            return [f'Walker {name} not found!']

    @interface.private_api()
    def walker_spawn_list(self, detailed: bool = False):
        """
        List walkers spawned by master
        """
        walks = []
        for i in self.spawned_walker_ids.obj_list():
            walks.append(i.serialize(detailed=detailed))
        return walks

    @interface.private_api(cli_args=['wlk'])
    def walker_prime(self, wlk: walker, nd: node = None, ctx: dict = {}):
        """
        Assigns walker to a graph node and primes walker for execution
        """
        wlk.prime(nd, prime_ctx=ctx)
        return [f'Walker primed on node {nd.id}']

    @interface.private_api(cli_args=['wlk'])
    def walker_execute(self, wlk: walker, prime: node = None,
                       ctx: dict = {}, profiling: bool = False):
        """
        Executes walker (assumes walker is primed)
        """
        if(prime):
            self.walker_prime(wlk=wlk, nd=prime, ctx=ctx)
        return wlk.run(profiling=profiling)

    @interface.private_api(cli_args=['name'])
    def walker_run(self, name: str, nd: node = None, ctx: dict = {},
                   snt: sentinel = None, profiling: bool = False):
        """
        Creates walker instance, primes walker on node, executes walker,
        reports results, and cleans up walker instance.
        """
        wlk = snt.spawn_walker(name, caller=self)
        if(not wlk):
            return self.bad_walk_response([f'Walker {name} not found!'])
        res = self.walker_execute(
            wlk=wlk, prime=nd, ctx=ctx, profiling=profiling)
        wlk.destroy()
        return res

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.spawned_walker_ids.obj_list():
            i.destroy()

    def bad_walk_response(self, errors=list()):
        return {'report': [], 'success': False, 'errors': errors}



"""
Super (master) api as a mixin
"""
from jaseci.api.interface import interface


class super_api():
    """Super APIs for creating nicknames for UUIDs and other long strings

    """

    @interface.admin_api(cli_args=['name'])
    def master_createsuper(self, name: str, set_active: bool = True,
                           other_fields: dict = {}):
        """
        Create a super instance and return root node super object

        other_fields used for additional feilds for overloaded interfaces
        (i.e., Dango interface)
        """
        from jaseci.element.super_master import super_master
        new_m = super_master(h=self._h, name=name)
        return self.make_me_head_master_or_destroy(new_m)
